# -*- coding: utf-8 -*-
import datetime
import re

import six
import pandas as pd

from rqdatac.client import get_client
from rqdatac.validators import (
    ensure_date_int,
    ensure_date_or_today_int,
    ensure_order_book_ids,
    ensure_string,
    ensure_string_in,
    check_items_in_container
)
from rqdatac.decorators import export_as_api
from rqdatac.utils import to_date_str


@export_as_api
def shenwan_industry(index_name, date=None, market="cn"):
    """获取申万行业组成

    :param index_name: 申万行业代码或名字, 如'801010.INDX'或'农林牧渔'
    :param date: 如 '2015-01-07' (Default value = None)
    :param market:  (Default value = "cn")
    :returns: 返回输入日期最近交易日的申万行业组成

    """
    if not isinstance(index_name, six.string_types):
        raise ValueError("string expected, got {!r}".format(index_name))

    if not date:
        date = datetime.date.today()
    date = ensure_date_int(date)
    return get_client().execute("shenwan_industry", index_name, date, market=market)


LEVEL_MAP = (
    None,
    ("index_code", "index_name"),
    ("second_index_code", "second_index_name"),
    ("third_index_code", "third_index_name"),
)


@export_as_api
def shenwan_instrument_industry(order_book_ids, date=None, level=1, expect_df=True, market="cn"):
    """获取股票对应的申万行业

    :param order_book_ids: 股票列表，如['000001.XSHE', '000002.XSHE']
    :param date: 如 '2015-01-07' (Default value = None)
    :param level:  (Default value = 1)
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :param market:  (Default value = "cn")
    :returns: code, name
        返回输入日期最近交易日的股票对应申万行业

    """

    if level not in [0, 1, 2, 3]:
        raise ValueError("level should be in 0,1,2,3")
    order_book_ids = ensure_order_book_ids(order_book_ids)

    if not date:
        date = datetime.date.today()
    date = ensure_date_int(date)

    r = get_client().execute("shenwan_instrument_industry", order_book_ids, date, level, market=market)
    if not r:
        return

    if len(order_book_ids) == 1 and not expect_df:
        r = r[0]
        if level != 0:
            return r[LEVEL_MAP[level][0]], r[LEVEL_MAP[level][1]]
        else:
            return (
                r["index_code"],
                r["index_name"],
                r["second_index_code"],
                r["second_index_name"],
                r["third_index_code"],
                r["third_index_name"],
            )

    df = pd.DataFrame(r).set_index("order_book_id")
    if level != 0 and level != 1:
        df.rename(columns=dict(zip(LEVEL_MAP[level], LEVEL_MAP[1])), inplace=True)
    return df


@export_as_api
def zx_industry(industry_name, date=None):
    """获取中信行业股票列表

    :param industry_name: 中信行业名称或代码
    :param date: 查询日期，默认为当前最新日期
    :return: 所属目标行业的order_book_id list or None
    """
    if not isinstance(industry_name, six.string_types):
        raise ValueError("string expected, got {!r}".format(industry_name))
    if not date:
        date = datetime.date.today()
    date = ensure_date_int(date)
    return get_client().execute("zx_industry", industry_name, date)


ZX_LEVEL_MAP = (
    None,
    "first_industry_name",
    "second_industry_name",
    "third_industry_name",
)


@export_as_api
def zx_instrument_industry(order_book_ids, date=None, level=1, expect_df=True):
    """获取股票对应的中信行业

    :param order_book_ids: 股票列表，如['000001.XSHE', '000002.XSHE']
    :param date: 如 '2015-01-07' (Default value = None)
    :param level:  (Default value = 1)
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :returns: code, name
        返回输入日期最近交易日的股票对应中信行业

    """

    if level not in [0, 1, 2, 3]:
        raise ValueError("level should be in 0,1,2,3")
    order_book_ids = ensure_order_book_ids(order_book_ids)

    if not date:
        date = datetime.date.today()
    date = ensure_date_int(date)

    r = get_client().execute("zx_instrument_industry", order_book_ids, date, level)
    if not r:
        return
    if len(order_book_ids) == 1 and not expect_df:
        r = r[0]
        if level != 0:
            return [r[ZX_LEVEL_MAP[level]], ]
        else:
            return [
                r["first_industry_name"],
                r["second_industry_name"],
                r["third_industry_name"],
            ]

    df = pd.DataFrame(r).set_index("order_book_id")
    return df


@export_as_api
def get_industry(industry, source='citics_2019', date=None, market="cn"):
    """获取行业股票列表

    :param industry: 行业名称或代码
    :param source: 分类来源。
                中国市场: citics 以及 citics_2019: 中信, gildata: 聚源
                港股: hsi: 恒生
    :param date: 查询日期，默认为当前最新日期
    :param market:  (Default value = "cn")
    :return: 所属目标行业的order_book_id list or None
    """

    industry = ensure_string(industry, "industry")
    source = ensure_string_in(source, ["sws", "citics", "gildata", "citics_2019", "hsi"], "source")
    date = ensure_date_or_today_int(date)

    res = get_client().execute("get_industry", industry, source, date, market=market)

    if not res:
        return res

    if res[-1] == "have_sector_name":
        # have_sector_name 代表 industry传入的是风格版块，产业板块或者上下游产业版块
        from rqdatac.services import basic
        res_list = basic.instruments(res[:-1])
        res = []
        date = to_date_str(date)
        for order_book in res_list:
            if order_book.de_listed_date == "0000-00-00" or order_book.de_listed_date is None:
                order_book.de_listed_date = "2099-12-31"
            if order_book.listed_date <= date <= order_book.de_listed_date:
                res.append(order_book.order_book_id)
    sub_pattern = re.compile('[A-Z]+')
    res = [sub_pattern.sub('', oid[:-4]) + oid[-4:] for oid in res]
    return sorted(res)


@export_as_api
def get_instrument_industry(order_book_ids, source='citics_2019', level=1, date=None, market="cn"):
    """获取股票对应的行业

    :param order_book_ids: 股票列表，如['000001.XSHE', '000002.XSHE']
    :param source: 分类来源。
                中国市场: citics 以及 citics_2019: 中信, gildata: 聚源
                港股: hsi: 恒生
    :param date: 如 '2015-01-07' (Default value = None)
    :param level:  (Default value = 1)
    :param market:  (Default value = "cn")
    :returns: code, name
        返回输入日期最近交易日的股票对应行业
    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    source = ensure_string_in(source, ["sws", "citics", "gildata", "citics_2019", "hsi", "sws_2014"], "source")
    if source == "citics_2019":
        check_items_in_container(level, [0, 1, 2, 3, "citics_sector"], 'level')
    else:
        check_items_in_container(level, [0, 1, 2, 3], 'level')
    date = ensure_date_or_today_int(date)

    r = get_client().execute("get_instrument_industry", order_book_ids, source, level, date, market=market)

    if not r:
        return
    res = [i['order_book_id'] for i in r]
    if source == "citics_2019" and level == "citics_sector":
        # is_special industry是否传入的是风格版块，产业板块和上下游产业版块
        from rqdatac.services import basic
        res_list = basic.instruments(res)
        date = to_date_str(date)
        for index, order_book in enumerate(res_list):
            if order_book.de_listed_date == "0000-00-00" or order_book.de_listed_date is None:
                order_book.de_listed_date = "2099-12-31"
            if not order_book.listed_date <= date <= order_book.de_listed_date:
                r.pop(index)

    return pd.DataFrame(r).set_index("order_book_id")


SHENWAN_COLUMNS = [
    "index_code",
    "index_name",
    "second_index_code",
    "second_index_name",
    "third_index_code",
    "third_index_name"
]
OTHER_COLUMNS = [
    "first_industry_code",
    "first_industry_name",
    "second_industry_code",
    "second_industry_name",
    "third_industry_code",
    "third_industry_name"
]


@export_as_api
def get_industry_mapping(source="citics_2019", date=None, market="cn"):
    """获取行业分类列表

    :param source: 分类来源。
                中国市场: citics 以及 citics_2019: 中信, gildata: 聚源
                港股: hsi: 恒生
    :param market:  (Default value = "cn")
    :return: DataFrame
    """
    source = ensure_string_in(source, ["sws", "citics", "gildata", "citics_2019", "hsi"], "source")
    if date is None:
        date = datetime.date.today()
    date = ensure_date_int(date)
    res = get_client().execute("get_industry_mapping_v2", source, date, market=market)
    if not res:
        return
    df = pd.DataFrame(res)

    if source == "sws":
        df.rename(columns=dict(zip(OTHER_COLUMNS, SHENWAN_COLUMNS)), inplace=True)
        columns = SHENWAN_COLUMNS
    else:
        columns = OTHER_COLUMNS

    df = df.dropna().drop_duplicates()
    df = df.sort_values(columns[::2]).reset_index(drop=True)
    return df[columns]


# -*- coding: utf-8 -*-
import struct
import sys
from datetime import datetime, timedelta
from decimal import Decimal
from functools import partial
from zlib import compress as zlib_compress, decompress as zlib_decompress

import msgpack

from .protocol import HEADER_FORMAT, COMPRESSION_METHOD, SERIALIZATION_TYPE


def raise_(e):
    raise e


_EPOCH = datetime(1970, 1, 1)


def _object_hook(obj):
    if "__$dt__" in obj:
        return _EPOCH + timedelta(seconds=obj["__$dt__"])
    if "__$dl__" in obj:
        return Decimal(obj["__$dl__"])
    return obj


def _default(obj):
    if isinstance(obj, datetime):
        return {"__$dt__": int((obj - _EPOCH).total_seconds())}
    if isinstance(obj, Decimal):
        return {"__$dl__": obj.to_eng_string()}
    return obj


use_bin_type = False if sys.version_info[0] == 2 else True
try:
    msgpack.loads(b'\xa4test', raw=True)
    msgpack_dumps = partial(msgpack.dumps, default=_default, use_bin_type=use_bin_type)
    msgpack_loads = partial(msgpack.loads, object_hook=_object_hook, raw=False,
                            max_str_len=2147483647,  # 2**32-1
                            max_bin_len=2147483647,
                            max_array_len=2147483647,
                            max_map_len=2147483647,
                            max_ext_len=2147483647
                            )
except TypeError:
    msgpack_dumps = partial(msgpack.dumps, default=_default, use_bin_type=use_bin_type)
    msgpack_loads = partial(msgpack.loads, object_hook=_object_hook, encoding="utf8")


try:
    import snappy

    snappy_decompress = snappy.decompress
    snappy_compress = snappy.compress
except (ImportError, AttributeError):
    snappy = None
    snappy_compress = snappy_decompress = lambda *_: raise_(
        RuntimeError(
            "compression not support {}".format(COMPRESSION_METHOD[COMPRESSION_METHOD.SNAPPY])
        )
    )

try:
    import brotli

    brotli_compress = partial(brotli.compress, quality=1)
    brotli_decompress = brotli.decompress
except ImportError:
    brotli = None
    brotli_compress = brotli_decompress = lambda *_: raise_(
        RuntimeError(
            "compression not support {}".format(COMPRESSION_METHOD[COMPRESSION_METHOD.BROTLI])
        )
    )


try:
    import rapidjson as json
except ImportError:
    import json
_json_dump = partial(json.dumps, default=_default)


def json_dumps(*args, **kwargs):
    return json.dumps(*args, default=_default, **kwargs).encode('utf-8')


json_loads = partial(json.loads, object_hook=_object_hook)


SERIALIZER = {
    SERIALIZATION_TYPE.RAW: lambda x: x,
    SERIALIZATION_TYPE.JSON: json_dumps,
    SERIALIZATION_TYPE.MSGPACK: msgpack_dumps,
}

DESERIALIZER = {
    SERIALIZATION_TYPE.RAW: lambda x: x,
    SERIALIZATION_TYPE.JSON: json_loads,
    SERIALIZATION_TYPE.MSGPACK: msgpack_loads,
}

COMPRESSION = {
    COMPRESSION_METHOD.NONE: lambda x: x,
    COMPRESSION_METHOD.SNAPPY: snappy_compress,
    COMPRESSION_METHOD.ZLIB: zlib_compress,
    COMPRESSION_METHOD.BROTLI: brotli_compress,
}

DECOMPRESSION = {
    COMPRESSION_METHOD.NONE: lambda x: x,
    COMPRESSION_METHOD.SNAPPY: snappy_decompress,
    COMPRESSION_METHOD.ZLIB: zlib_decompress,
    COMPRESSION_METHOD.BROTLI: brotli_decompress,
}


_header_struct = struct.Struct(HEADER_FORMAT)
pack_header = _header_struct.pack
unpack_header = _header_struct.unpack


def pack_one(data, mt, st, cm, force_compress=False):
    # type: (...,int, int, int) -> bytes
    data = SERIALIZER[st](data)
    if force_compress or len(data) > 1024:
        data = COMPRESSION[cm](data)
    else:
        cm = COMPRESSION_METHOD.NONE
    return pack_header(mt, st, cm, len(data)) + data


def unpack_one(data, st, cm):
    return DESERIALIZER[st](DECOMPRESSION[cm](data))



# -*- coding: utf-8 -*-
import bisect
import datetime
from rqdatac.client import get_client
from rqdatac.utils import int8_to_date, int_to_datetime
from rqdatac.validators import ensure_date_int
from rqdatac.decorators import export_as_api, ttl_cache, compatible_with_parm


@ttl_cache(24 * 3600)
def _get_all_trading_dates(market):
    return get_client().execute("get_all_trading_dates", market=market)


def _map_expect_type(ty, fmt, dates):
    if ty == "int":
        return dates
    if ty == "datetime":
        return [int_to_datetime(dt) for dt in dates]
    if ty == "date":
        return [int8_to_date(dt) for dt in dates]
    if ty == "str":
        return [int_to_datetime(dt).strftime(fmt) for dt in dates]
    raise TypeError(ty)


def _expect_type(ty, fmt, date):
    return _map_expect_type(ty, fmt, [date])[0]


def get_trading_dates_in_type(start_date, end_date, expect_type="datetime", fmt=None, market="cn"):
    """获取两个日期之间的交易日列表

    :param start_date: 开始日期
    :param end_date: 结束日期
    :param expect_type:  (Default value = "datetime")
    :param fmt:  (Default value = None)
    :param market:  (Default value = "cn")

    """
    start_date = ensure_date_int(start_date)
    end_date = ensure_date_int(end_date)
    dates = _get_all_trading_dates(market)
    start_pos = bisect.bisect_left(dates, start_date)
    end_pos = bisect.bisect_right(dates, end_date)
    return _map_expect_type(expect_type, fmt, dates[start_pos:end_pos])


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def get_trading_dates(start_date, end_date, market="cn"):
    """获取两个日期之间的交易日列表

    :param start_date: 如 '2013-01-04'
    :param end_date: 如 '2013-01-04'
    :param market: 地区代码, 如 'cn' (Default value = "cn")
    :returns: 日期列表

    """
    start_date = ensure_date_int(start_date)
    end_date = ensure_date_int(end_date)
    dates = _get_all_trading_dates(market)
    start_pos = bisect.bisect_left(dates, start_date)
    end_pos = bisect.bisect_right(dates, end_date)
    return [int8_to_date(i) for i in dates[start_pos:end_pos]]


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def get_next_trading_date(date, n=1, market="cn"):
    """获取后一交易日

    :param date: 日期
    :parm n: 日期间隔
    :param market:  (Default value = "cn")

    """
    if n < 1:
        raise ValueError("n: except a positive value, got {}".format(n))
    date = ensure_date_int(date)
    dates = _get_all_trading_dates(market)
    pos = bisect.bisect_right(dates, date)
    if pos + n - 1 < len(dates):
        return int8_to_date(dates[pos + n - 1])
    return int8_to_date(dates[-1])


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def get_previous_trading_date(date, n=1, market="cn"):
    """获取前一交易日

    :param date: 日期
    :parm n: 日期间隔
    :param market:  (Default value = "cn")

    """
    if n < 1:
        raise ValueError("n: except a positive value, got {}".format(n))
    date = ensure_date_int(date)
    dates = _get_all_trading_dates(market)
    pos = bisect.bisect_left(dates, date)
    if pos > n:
        return int8_to_date(dates[pos - n])
    return int8_to_date(dates[0])


@export_as_api
def get_latest_trading_date(market="cn"):
    """获取最近的一个交易日

    :param market:  (Default value = "cn")

    """
    tomorrow = datetime.date.today() + datetime.timedelta(days=1)
    return get_previous_trading_date(tomorrow, market=market)


@export_as_api
def trading_date_offset(date, offset, market="cn"):
    """获取当前日期之前（或之后）的一个交易日

    :param date: 日期
    :param offset: 日期间隔
    :param market:  (Default value = "cn")

    """
    date = ensure_date_int(date)
    dates = _get_all_trading_dates(market)
    if offset < 0:
        pos = bisect.bisect_left(dates, date) + offset
    else:
        pos = bisect.bisect_right(dates, date) + offset

    return int8_to_date(dates[pos])


@export_as_api
def is_trading_date(date, market="cn"):
    """判断日期是否为交易日
    :param date: 日期 如20190401
    :param market:  (Default value = "cn")
    :returns: bool
    """
    date = ensure_date_int(date)
    dates = _get_all_trading_dates(market)
    return date in dates


def get_prev_weekday(date):
    """获取上一个工作日，周二到周四返回昨天，周一返回上周五
    :param date : date_int 如 如20190411
    :returns: date_int 如20190413
    """
    date = int8_to_date(date)
    weekday = date.weekday()
    if weekday == 0:
        days = 3
    else:
        days = 1
    return date - datetime.timedelta(days=days)


@export_as_api
def has_night_trading(date, market="cn"):
    """判断交易日是否有夜盘
    :param date: 日期 如20190401
    :param market:  (Default value = "cn")
    :returns: bool
    """
    date = ensure_date_int(date)
    if not is_trading_date(date, market):
        return False
    return is_trading_date(get_prev_weekday(date))



# -*- coding: utf-8 -*-
import warnings
import datetime

import pandas as pd
import numpy as np
from rqdatac.services.constant import RATETYPE_CN, RATECOMP_CN

from rqdatac.client import get_client
from rqdatac.validators import (
    ensure_int,
    ensure_date_int,
    ensure_order_book_id,
    ensure_order_book_ids,
    ensure_date_range,
    ensure_dates_base_on_listed_date,
    ensure_list_of_string, ensure_date_or_today_int, check_items_in_container)
from rqdatac.utils import to_datetime, int8_to_datetime
from rqdatac.decorators import export_as_api, ttl_cache
from rqdatac.services.calendar import (
    get_trading_dates,
)
from rqdatac.services import shenwan

INS_COLUMNS = [
    "order_book_id",
    "symbol",
    "full_name",
    "exchange",
    "bond_type",
    "trade_type",
    "value_date",
    "maturity_date",
    "par_value",
    "coupon_rate",
    "coupon_frequency",
    "coupon_method",
    "compensation_rate",
    "total_issue_size",
    "de_listed_date",
    "stock_code",
    "conversion_start_date",
    "conversion_end_date",
    "redemption_price",
    "issue_price",
    "call_protection",
    "listed_date",
    "stop_trading_date"
]


class Instrument:
    def __init__(self, attrs):
        self.__dict__.update(attrs)
        self.__cache = {}

    def __str__(self):
        return "{}(\n{}\n)".format(
            type(self).__name__,
            ",\n".join(["{}={!r}".format(k, v) for k, v in self.items() if not k.startswith("_")]),
        )

    __repr__ = __str__

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, item):
        return self.__dict__[item]

    def get(self, item, default=None):
        return self.__dict__.get(item, default)

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def __cache_get(self, v):
        return self.__cache.get(v)

    def __cache_set(self, k, v):
        self.__cache[k] = v

    def coupon_rate_table(self):
        """变动利率可转债信息"""
        if "coupon_rate_table" in self.__cache:
            return self.__cache_get("coupon_rate_table")
        info = get_client().execute("convertible.get_coupon_rate_table", self.order_book_id)
        info = pd.DataFrame(info).set_index(['start_date', 'end_date']) if info else None
        self.__cache_set("coupon_rate_table", info)
        return info

    def option(self, option_type=None):
        if option_type is not None:
            option_type = ensure_int(option_type)
            if option_type not in (1, 2, 3, 4, 5, 6, 7):
                raise ValueError("option_type: expect value in (None, 1, 2, 3, 4, 5, 6, 7)")

        data = get_client().execute("convertible.option", self.order_book_id, option_type)
        if not data:
            return

        df = pd.DataFrame(data)
        if 'payment_year' in df.columns:
            sort_fields = ['option_type', 'payment_year']
        else:
            sort_fields = ['option_type']
        df = df.sort_values(sort_fields).reset_index()
        column_order = ['option_type', 'start_date', 'end_date', 'payment_year', 'level', 'window_days',
                        'reach_days', 'frequency', 'price', 'if_include_interest', 'remark']
        column = [i for i in column_order if i in df.columns]
        return df[column]


@ttl_cache(12 * 3600)
def _all_instruments_dict(market="cn"):
    return {
        i['order_book_id']: Instrument(i)
        for i in get_client().execute("convertible.all_instruments", market=market)
    }


@export_as_api(namespace="convertible")
def all_instruments(date=None, market="cn"):
    """获取所有可转债详细信息

    :param market:  (Default value = "cn")
    :returns: DataFrame
    """
    profile = lambda v: (
        v.order_book_id,
        v.symbol,
        v.full_name,
        v.exchange,
        v.bond_type,
        v.trade_type,
        v.value_date,
        v.maturity_date,
        v.par_value,
        v.coupon_rate,
        v.coupon_frequency,
        v.coupon_method,
        v.compensation_rate,
        v.total_issue_size,
        v.de_listed_date,
        v.stock_code,
        v.conversion_start_date,
        v.conversion_end_date,
        v.redemption_price,
        v.issue_price,
        v.call_protection,
        v.listed_date,
        getattr(v, 'stop_trading_date', None)
    )

    def judge(listed_date, de_listed_date):
        if listed_date and de_listed_date:
            return listed_date <= date and de_listed_date > date
        if listed_date:
            return listed_date <= date
        else:
            return False

    if date:
        date = to_datetime(date)
        data = [profile(v) for v in _all_instruments_dict(market).values() if judge(v.listed_date, v.de_listed_date)]
    else:
        data = [profile(v) for v in _all_instruments_dict(market).values()]
    df = pd.DataFrame(
        data,
        columns=INS_COLUMNS,
    )
    df.sort_values('order_book_id', inplace=True)
    return df.reset_index(drop=True)


@export_as_api(namespace="convertible")
def instruments(order_book_ids, market="cn"):
    """获取可转债详细信息

    :param order_book_ids: 可转债代码，str 或 list of str
    :param market:  (Default value = "cn")
    :returns: Instrument object or list of Instrument object
            取决于参数是一个 order_book_id 还是多个 order_book_id
    """
    all_dict = _all_instruments_dict(market)
    order_book_ids = ensure_list_of_string(order_book_ids, "order_book_ids")
    if len(order_book_ids) == 1:
        try:
            return all_dict[order_book_ids[0]]
        except KeyError:
            warnings.warn('unknown convertible order_book_id: {}'.format(order_book_ids))
            return
    all_list = (all_dict.get(i) for i in order_book_ids)
    return [i for i in all_list if i]


@export_as_api(namespace="convertible")
def get_cash_flow(order_book_ids, start_date=None, end_date=None, market="cn"):
    """获取现金流信息

    :param order_book_ids: 可转债ID str or list
    :param start_date: 开始日期，默认为None
    :param end_date: 结束日期，默认为None
    :param market:  (Default value = "cn")
    :return: pd.DataFrame
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    if start_date:
        start_date = ensure_date_int(start_date)
    if end_date:
        end_date = ensure_date_int(end_date)
    data = get_client().execute("convertible.get_cash_flow", order_book_ids, start_date, end_date, market=market)
    if not data:
        return
    df = pd.DataFrame(data)
    df.set_index(["order_book_id", "payment_date"], inplace=True)
    return df


@export_as_api(namespace="convertible")
def get_call_info(order_book_ids, start_date=None, end_date=None, market="cn"):
    """获取赎回信息

    :param order_book_ids: 可转债ID，str or list
    :param start_date: 开始日期，默认为None
    :param end_date: 结束日期，默认为None
    :param market:  (Default value = "cn")
    :return: pd.DataFrame
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    if start_date:
        start_date = ensure_date_int(start_date)
    if end_date:
        end_date = ensure_date_int(end_date)
    data = get_client().execute("convertible.get_call_info", order_book_ids, start_date, end_date, market=market)
    if not data:
        return
    df = pd.DataFrame(data)
    df.set_index(["order_book_id", "info_date"], inplace=True)
    return df


@export_as_api(namespace="convertible")
def get_put_info(order_book_ids, start_date=None, end_date=None, market="cn"):
    """获取回售信息

    :param order_book_ids: 可转债ID，str or list
    :param start_date: 开始日期，默认为None
    :param end_date: 结束日期，默认为None
    :param market:  (Default value = "cn")
    :return: pd.DataFrame
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    if start_date:
        start_date = ensure_date_int(start_date)
    if end_date:
        end_date = ensure_date_int(end_date)
    data = get_client().execute("convertible.get_put_info", order_book_ids, start_date, end_date, market=market)
    if not data:
        return
    df = pd.DataFrame(data)
    df.set_index(["order_book_id", "info_date"], inplace=True)
    return df


@export_as_api(namespace="convertible")
def get_conversion_price(order_book_ids, start_date=None, end_date=None, market="cn"):
    """获取转股价变动信息

    :param order_book_ids: 可转债ID，str or list
    :param start_date: 开始日期，默认为None
    :param end_date: 结束日期，默认为None
    :param market:  (Default value = "cn")
    :return: pd.DataFrame
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    if start_date:
        start_date = ensure_date_int(start_date)
    if end_date:
        end_date = ensure_date_int(end_date)
    data = get_client().execute("convertible.get_conversion_price", order_book_ids, start_date, end_date, market=market)
    if not data:
        return
    df = pd.DataFrame(data)
    df.set_index(["order_book_id", "info_date"], inplace=True)
    return df


@export_as_api(namespace="convertible")
def get_conversion_info(order_book_ids, start_date=None, end_date=None, market="cn"):
    """获取转股变动信息

    :param order_book_ids: 可转债ID，str or list
    :param start_date: 开始日期，默认为None
    :param end_date: 结束日期，默认为None
    :param market:  (Default value = "cn")
    :return: pd.DataFrame
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    if start_date:
        start_date = ensure_date_int(start_date)
    if end_date:
        end_date = ensure_date_int(end_date)
    data = get_client().execute("convertible.get_conversion_info", order_book_ids, start_date, end_date, market=market)
    if not data:
        return
    df = pd.DataFrame(data)
    df.set_index(["order_book_id", "info_date"], inplace=True)
    return df


@export_as_api(namespace="convertible")
def is_suspended(order_book_ids, start_date=None, end_date=None):
    """获取停牌信息
    :param order_book_ids: 可转债ID
    :param start_date: 开始日期, 如'2013-01-04' (Default value = None)
    :param end_date: 结束日期，如'2014-01-04' (Default value = None)
    :returns: DataFrame
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    if len(order_book_ids) == 1:
        instrument = instruments(order_book_ids[0], market="cn")
        start_date, end_date = ensure_dates_base_on_listed_date(instrument, start_date, end_date, "cn")
        if start_date is None:
            return
    start_date, end_date = ensure_date_range(start_date, end_date)

    trading_dates = pd.to_datetime(get_trading_dates(start_date, end_date, market="cn"))
    df = pd.DataFrame(data=False, columns=order_book_ids, index=trading_dates)
    data = get_client().execute("convertible.is_suspended", order_book_ids, start_date, end_date, market="cn")
    for (order_book_id, date) in data:
        date = to_datetime(date)
        df.at[date, order_book_id] = True
    return df


ISSUER_FIELDS = [
    'credit_date',
    'credit_level',
    'rating_type',
    'institution',
    'organization',
    'org_type',
    'outlook',
    'adjusted_direction',
]
DEBT_FIELDS = [
    'credit_date',
    'credit_level',
    'rating_type',
    'institution',
    'outlook',
    'adjusted_direction',
]


@export_as_api(namespace="convertible")
def get_credit_rating(order_book_ids, start_date=None, end_date=None, institution=None, target='debt', fields=None):
    """
    :param order_book_ids: bond id, str or str list
    :param start_date: start date, datatime, int or str
    :param end_date: end date, datatime, int or str
    :param institution: crediting company name
    :param target: 'debt' or 'issuer'
    :param fields: 筛选字段
            debt: [
                'credit_date','credit_level','rating_type','institution','outlook','adjusted_direction'
            ]
            issuer: [
                'credit_date','credit_level','rating_type','institution',
                'organization', 'org_type','outlook','adjusted_direction',
            ]
    :return: a pandas DataFrame with order_book_id as index
    """
    check_items_in_container(target, ['debt', 'issuer'], 'target')

    all_fields = DEBT_FIELDS if target == 'debt' else ISSUER_FIELDS
    fields = ensure_list_of_string(fields, "fields") if fields else all_fields
    order_book_ids = ensure_list_of_string(order_book_ids, "order_book_ids")
    start_date = ensure_date_int(start_date) if start_date is not None else start_date
    end_date = ensure_date_int(end_date) if end_date is not None else end_date
    if institution == "中诚信证券评估有限公司":
        institution = "中诚信证评数据科技有限公司"
    data = get_client().execute(
        'convertible.get_credit_rating', order_book_ids, start_date, end_date, institution, target, fields
    )
    if not data:
        return
    df = pd.DataFrame(data)
    df.set_index(['order_book_id'], inplace=True)
    return df.reindex(columns=fields)


@export_as_api(namespace="convertible")
def rating(date=None, credit_level=None, institution=None, rating_type=None, target='debt'):
    """
    Get rating information for company or bond
    :param date: str, int, or datatime
        1): 存续债券的判定是date在[value_date,maturity_date]之间
        2): 控制credit_date返回当前最新的日期
    :param credit_level: eg: 'AAA'
    :param institution: rating company name
    :param rating_type: rating type name
    :param target: 'debt' or 'issuer'
    :return:
    """
    check_items_in_container(target, ['debt', 'issuer'], 'target')
    if institution is not None:
        check_items_in_container(institution, RATECOMP_CN, 'institution')
        if institution == "中诚信证券评估有限公司":
            institution = "中诚信证评数据科技有限公司"
    if rating_type is not None:
        check_items_in_container(rating_type, RATETYPE_CN, 'rating_type')
    date = ensure_date_int(date) if date else None

    res = get_client().execute("convertible.rating", date, credit_level, institution, rating_type, target)

    if date and res and target == 'debt':
        ins = instruments(res)
        if not isinstance(ins, list):
            ins = [ins]
        res = [i.order_book_id for i in ins if i.value_date <= int8_to_datetime(date) <= i.maturity_date]
    return res


@export_as_api(namespace="convertible")
def get_latest_rating(order_book_ids, date, institution=None, rating_type=None, target='debt'):
    """
    获取在给定日期之前的最新评级记录.
    返回credit_date和参数date前差距时间最短的一条记录，无需返回所有评级机构最新记录

    :param order_book_ids: str or List[str]债券id列表
    :param date: str or int or datetime.date) 评级日期; 会返回该日期之前的最新评级记录.
    :param institution: str or List[str] or None 评级机构; 若为None, 则返回所有评级机构的最新记录
    :param rating_type: str or None 评级类型; 如果给定的话, 只返回该评级类型下最新的评级信息,
        如果设为None, 则不管评级类型, 直接返回最新评级记录
    :param target: str 评级类型, 可选值为 'debt'(代表债券评级) 或者 'issuer'(代表主体评级);
    :return: a pandas DataFrame with order_book_id as index.
    """
    order_book_ids = ensure_list_of_string(order_book_ids, "order_book_ids")
    check_items_in_container(target, ['debt', 'issuer'], 'target')
    if institution is not None:
        institution = ensure_list_of_string(institution, "institution")
        check_items_in_container(institution, RATECOMP_CN, 'institution')
        if "中诚信证券评估有限公司" in institution:
            institution.append("中诚信证评数据科技有限公司")

    data = get_client().execute(
        "convertible.get_latest_rating",
        order_book_ids,
        ensure_date_int(date),
        institution,
        rating_type,
        target
    )
    if not data:
        return
    df = pd.DataFrame.from_records(data)
    df.sort_values(["order_book_id", "credit_date"], ascending=False, inplace=True)
    df.set_index("order_book_id", inplace=True)
    return df


@export_as_api(namespace="convertible")
def get_instrument_industry(order_book_ids, source='citics', level=1, date=None, market="cn"):
    """获取可转债对应的行业

    :param order_book_ids: 可转债order_book_id，如['000001.XSHE', '000002.XSHE']
    :param source: 分类来源。citics 以及 citics_2019: 中信, gildata: 聚源
    :param date: 如 '2015-01-07' (Default value = None)
    :param level:  (Default value = 1)
    :param market:  (Default value = "cn")
    :returns: DataFrame
        index: order_book_id
        columns:
            if level == 1: ["first_industry_code", "first_industry_name"]
            if level == 2: ["second_industry_code", "second_industry_name"]
            if level == 3: ["third_industry_code", "third_industry_name"]
            if level == 0:
                [
                    "first_industry_code", "first_industry_name", "second_industry_code",
                    "second_industry_name", "third_industry_code", "third_industry_name",
                ]
    """
    order_book_ids = ensure_order_book_ids(order_book_ids, type="Convertible")
    all_dict = _all_instruments_dict(market)
    stock_code_map = {
        all_dict.get(i).stock_code: all_dict.get(i).order_book_id
        for i in order_book_ids if all_dict.get(i)
    }
    # 调用股票行业接口
    res = shenwan.get_instrument_industry(list(stock_code_map.keys()), source, level, date, market)
    if res is None:
        return

    # 转换order_book_id为可转债id
    res.index = res.index.map(stock_code_map)
    return res


@export_as_api(namespace="convertible")
def get_industry(industry, source='citics', date=None, market="cn"):
    """获取行业可转债列表

    :param industry: 行业名称或代码
    :param source: 分类来源。citics 以及 citics_2019: 中信, gildata: 聚源
    :param date: 查询日期，默认为当前最新日期
    :param market:  (Default value = "cn")
    :return: 所属目标行业的order_book_id list or None
    """
    # 调用股票行业接口
    order_book_ids = shenwan.get_industry(industry, source, date, market)
    if order_book_ids is None:
        return

    order_book_ids = set(order_book_ids)

    all_dict = _all_instruments_dict(market)
    if date:
        oids = []
        date = to_datetime(date)
        for ins in all_dict.values():
            if ins.stock_code in order_book_ids:
                if ins.de_listed_date == "0000-00-00" or ins.de_listed_date is None:
                    ins.de_listed_date = pd.to_datetime("2099-12-31")
                if ins.listed_date is None:
                    ins.listed_date = pd.to_datetime("2099-12-31")
                if ins.listed_date <= date <= ins.de_listed_date:
                    oids.append(ins.order_book_id)
    else:
        oids = [ins.order_book_id for ins in all_dict.values() if ins.stock_code in order_book_ids]
    return sorted(oids)


@export_as_api(namespace="convertible")
def get_indicators(order_book_ids, start_date=None, end_date=None, fields=None):
    """获取可转债指标(默认返回最新3个月的数据)

    :param order_book_ids: str or List[str] 合约代码
    :param start_date: (str or datetime.date): 开始日期
    :param end_date: (str or datetime.date): 结束时间
    :param fields: str or List[str] 筛选字段

    :return: Multi-index DataFrame:
        index: [order_book_id, date]
        columns:
            conversion_coefficient	                float	转股系数
            conversion_value	                    float	转股价值
            conversion_premium	                    float	转股溢价率
            pure_bond_value_premium	                float	税后纯债溢价率
            pure_bond_value_premium_pretax	        float   税前纯债溢价率
            yield_to_maturity	                    float	税后到期收益率
            yield_to_maturity_pretax	            float   税前到期收益率
            yield_to_put	                        float	税后回售收益率
            yield_to_put_pretax	                    float	税前回售收益率
            pure_bond_value	                        float	税后纯债价值
            pure_bond_value_pretax	                float	税前纯债价值
            double_low_factor	                    float	双低指标
            call_trigger_price	                    float	赎回触发价
            put_trigger_price	                    float	回售触发价
            conversion_price_reset_trigger_price	float	下修触发价
            turnover_rate	                        float 	换手率
            remaining_size	                        float 	剩余规模（元）
            convertible_market_cap_ratio	        float 	转债市值占比
            pb_ratio	                            float	市净率
            put_qualified_days	                    float	回售已满足天数
            call_qualified_days	                    float	赎回已满足天数
            conversion_price_reset_qualified_days	float	转股价下修已满足天数
            put_status	                            float	回售条款满足状态
            call_status	                            float	强赎条款满足状态
            conversion_price_reset_status	        float	下修条款满足状态
    """
    all_fields = [
        "conversion_coefficient", "conversion_value", "conversion_premium", "pure_bond_value_premium",
        "pure_bond_value_premium_pretax", "yield_to_maturity", "yield_to_maturity_pretax", "yield_to_put",
        "yield_to_put_pretax", "pure_bond_value", "pure_bond_value_pretax", "double_low_factor",
        "call_trigger_price", "put_trigger_price", "conversion_price_reset_trigger_price", "turnover_rate",
        "remaining_size", "convertible_market_cap_ratio", "pb_ratio", "put_qualified_days",
        "call_qualified_days", "conversion_price_reset_qualified_days", "put_status", "call_status",
        "conversion_price_reset_status"
    ]
    order_book_ids = ensure_list_of_string(order_book_ids, "order_book_ids")
    if fields is None:
        fields = all_fields
    else:
        fields = ensure_list_of_string(fields, "fields")
        check_items_in_container(fields, all_fields, 'fields')

    # 默认返回最新3个月的数据
    if start_date is None and end_date is None:
        start_date, end_date = ensure_date_range(start_date, end_date)
    start_date = ensure_date_int(start_date) if start_date is not None else start_date
    end_date = ensure_date_int(end_date) if end_date is not None else end_date

    data = get_client().execute(
        "convertible.get_indicators", order_book_ids, start_date, end_date, fields
    )
    if not data:
        return

    data = pd.DataFrame(data)
    data.set_index(["order_book_id", "date"], inplace=True)
    data.sort_index(inplace=True)
    return data


@export_as_api(namespace="convertible")
def get_coupon_rate_table(order_book_ids):
    """ 变动利率可转债信息

    :param order_book_ids: str or List[str] 合约代码
    :return: DataFrame
        index: ['order_book_id', 'start_date', 'end_date']
        columns: ['coupon_rate']
    """
    order_book_ids = ensure_list_of_string(order_book_ids, "order_book_ids")
    info = get_client().execute("convertible.get_coupon_rate_tables", order_book_ids)
    info = pd.DataFrame(info).set_index(['order_book_id', 'start_date', 'end_date']) if info else None

    return info


@export_as_api(namespace="convertible")
def get_accrued_interest_eod(order_book_ids, start_date=None, end_date=None):
    """ 获取可转债应计利息

    只输入开始日期，返回开始日往后3个月内数值；
    只输入结束日期，返回结束日往前3个月应计利息；
    不输入日期，则返回当前日期往以前3个月数值

    :param order_book_ids: str or List[str] 合约代码
    :param start_date: 开始时间
    :param end_date: 结束时间
    :return: DataFrame
        index: date
        columns: order_book_ids      str or List[str]
    """
    order_book_ids = ensure_list_of_string(order_book_ids, "order_book_ids")

    if start_date is None or end_date is None:
        start_date, end_date = ensure_date_range(start_date, end_date)
    start_date = to_datetime(start_date)
    end_date = to_datetime(end_date)

    ins = instruments(order_book_ids)
    if ins is None:
        return

    ins = ins if isinstance(ins, list) else [ins]
    # 去掉当天已经完成赎回的转债
    order_book_ids = [c.order_book_id for c in ins if c.bond_type != "separately_traded" and c.maturity_date > start_date]
    if not order_book_ids:
        return None

    # 获得强赎信息
    called_info = get_call_info(order_book_ids)

    # 获取coupon_rate_table
    coupon_rate_tables = get_coupon_rate_table(order_book_ids)
    coupon_rate_tables.reset_index(inplace=True)

    res_list = []
    for oid, sub_df in coupon_rate_tables.groupby("order_book_id"):
        record_date = sub_df.iloc[-1].end_date
        if called_info is not None and oid in called_info.index.get_level_values(0):
            # 获取登记时间
            record_date = called_info.loc[oid].record_date[0]

        # 生成时间段内对应的coupon_rate
        df = pd.DataFrame(columns=["end_date", "date"])
        date_range = pd.date_range(sub_df.iloc[0].start_date, record_date)
        df.end_date, df.date = date_range, date_range
        df = df.merge(sub_df, on="end_date", how="left")
        df = df.bfill()
        df.dropna(inplace=True)

        # 计算利息
        df["value"] = (df.date - df.start_date) / datetime.timedelta(days=365) * df.coupon_rate * 100

        # 闰日用前一天的利息补
        # FIXME 闰日之后的所有利息需往前挪一天
        df.loc[(df["date"].dt.month == 2) & (df["date"].dt.day == 29), "value"] = np.nan
        df = df.ffill()

        df = df[(df.date >= start_date) & (df.date <= end_date)]
        df = df.pivot(index="date", columns="order_book_id", values="value")
        df.columns.name = None
        df = df.sort_index()
        res_list.append(df)

    res = pd.concat(res_list, axis=1) if res_list else None
    return res



# -*- coding: utf-8 -*-
import six
import datetime

import pandas as pd
from dateutil.relativedelta import relativedelta
from rqdatac.validators import (
    ensure_string,
    ensure_list_of_string,
    ensure_date_int,
    ensure_date_or_today_int,
    ensure_date_range,
    check_items_in_container,
)
from rqdatac import get_trading_dates
from rqdatac.client import get_client
from rqdatac.decorators import export_as_api, ttl_cache
from rqdatac.utils import int8_to_datetime, to_datetime


@export_as_api
def get_dominant_future(underlying_symbol, start_date=None, end_date=None, rule=0, rank=1, market="cn"):
    import warnings

    msg = "'get_dominant_future' is deprecated, please use 'futures.get_dominant' instead"
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    return get_dominant(underlying_symbol, start_date, end_date, rule, rank, market)


@export_as_api(namespace='futures')
def get_dominant(underlying_symbol, start_date=None, end_date=None, rule=0, rank=1, market="cn"):
    """获取指定期货品种当日对应的主力合约

    :param underlying_symbol: 如'IF' 'IC'
    :param start_date: 如 '2015-01-07' (Default value = None)
    :param end_date: 如 '2015-01-08' (Default value = None)
    :param market:  (Default value = "cn")
    :param rule:  主力合约规则 (Default value = 0)
        0：在rule=1的规则上，增加约束(曾做过主力合约的合约，一旦被换下来后，不会再被选上)
        1：合约首次上市时，以当日收盘同品种持仓量最大者作为从第二个交易日开始的主力合约，当同品种其他合约持仓量在收盘后
           超过当前主力合约1.1倍时，从第二个交易日开始进行主力合约的切换。日内不会进行主力合约的切换
    :param rank:  (Default value = 1):
        1: 主力合约
        2: 次主力合约
        3：次次主力合约
    :returns: pandas.Series
        返回参数指定的具体主力合约名称

    """
    if not isinstance(underlying_symbol, six.string_types):
        raise ValueError("invalid underlying_symbol: {}".format(underlying_symbol))

    check_items_in_container(rule, [0, 1], 'rule')
    check_items_in_container(rank, [1, 2, 3], 'order')

    underlying_symbol = underlying_symbol.upper()

    if start_date:
        start_date = ensure_date_int(start_date)

    if end_date:
        end_date = ensure_date_int(end_date)
    elif start_date:
        end_date = start_date

    if rank == 1:
        result = get_client().execute(
            "futures.get_dominant", underlying_symbol, start_date, end_date, rule, market=market)
    else:
        result = get_client().execute(
            "futures.get_dominant_v2", underlying_symbol, start_date, end_date, rule, rank, market=market)

    if not result:
        return
    df = pd.DataFrame(result)
    df["date"] = df["date"].apply(int8_to_datetime)
    return df.set_index("date").sort_index()["dominant"]


_FIELDS = [
    "margin_type",
    "long_margin_ratio",
    "short_margin_ratio",
    "commission_type",
    "open_commission_ratio",
    "close_commission_ratio",
    "close_commission_today_ratio",
]


@export_as_api
def future_commission_margin(order_book_ids=None, fields=None, hedge_flag="speculation"):
    import warnings

    msg = "'future_commission_margin' is deprecated, please use 'futures.get_commission_margin' instead"
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    return get_commission_margin(order_book_ids, fields, hedge_flag)


@export_as_api(namespace='futures')
def get_commission_margin(order_book_ids=None, fields=None, hedge_flag="speculation"):
    """获取期货保证金和手续费数据

    :param order_book_ids: 期货合约, 支持 order_book_id 或 order_book_id list,
        若不指定则默认获取所有合约 (Default value = None)
    :param fields: str 或 list, 可选字段有： 'margin_type', 'long_margin_ratio', 'short_margin_ratio',
            'commission_type', 'open_commission_ratio', 'close_commission_ratio',
            'close_commission_today_ratio', 若不指定则默认获取所有字段 (Default value = None)
    :param hedge_flag: str, 账户对冲类型, 可选字段为: 'speculation', 'hedge',
            'arbitrage', 默认为'speculation', 目前仅支持'speculation' (Default value = "speculation")
    :returns: pandas.DataFrame

    """
    if order_book_ids:
        order_book_ids = ensure_list_of_string(order_book_ids)

    if fields is None:
        fields = _FIELDS
    else:
        fields = ensure_list_of_string(fields, "fields")
        check_items_in_container(fields, _FIELDS, "fields")

    hedge_flag = ensure_string(hedge_flag, "hedge_flag")
    if hedge_flag not in ["speculation", "hedge", "arbitrage"]:
        raise ValueError("invalid hedge_flag: {}".format(hedge_flag))

    ret = get_client().execute("futures.get_commission_margin", order_book_ids, fields, hedge_flag)
    return pd.DataFrame(ret)


@export_as_api
def get_future_member_rank(order_book_id, trading_date=None, info_type='volume'):
    import warnings

    msg = "'get_future_member_rank' is deprecated, please use 'futures.get_member_rank' instead"
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    return get_member_rank(order_book_id, trading_date, info_type)


@export_as_api(namespace='futures')
def get_member_rank(obj, trading_date=None, rank_by='volume', **kwargs):
    """获取指定日期最近的期货会员排名数据
    :param obj： 期货合约或品种代码
    :param trading_date: 日期
    :param rank_by: 排名依据字段
    :keyword start_date
    :keyword end_date
    :returns pandas.DataFrame or None
    """
    if not kwargs:
        trading_date = ensure_date_or_today_int(trading_date)
        ret = get_client().execute("futures.get_member_rank", obj, trading_date, rank_by)
    else:
        start_date = kwargs.pop("start_date", None)
        end_date = kwargs.pop("end_date", None)
        if kwargs:
            raise ValueError('unknown kwargs: {}'.format(kwargs))
        elif start_date and end_date:
            start_date, end_date = ensure_date_int(start_date), ensure_date_int(end_date)
            ret = get_client().execute("futures.get_member_rank_v2", obj, start_date, end_date, rank_by)
        else:
            raise ValueError('please ensure start_date and end_date exist')

    if not ret:
        return

    df = pd.DataFrame(ret).sort_values(by=['trading_date', 'rank'])
    df.set_index('trading_date', inplace=True)
    return df


@export_as_api(namespace="futures")
def get_warehouse_stocks(underlying_symbols, start_date=None, end_date=None, market="cn"):
    """获取时间区间内期货的注册仓单

    :param underlying_symbols: 期货品种, 支持列表查询
    :param start_date: 如'2015-01-01', 如果不填写则为去年的当日日期
    :param end_date: 如'2015-01-01', 如果不填写则为当日日期
    :param market: 市场, 默认为"cn"
    :return: pd.DataFrame

    """
    underlying_symbols = ensure_list_of_string(underlying_symbols, name="underlying_symbols")
    start_date, end_date = ensure_date_range(start_date, end_date, delta=relativedelta(years=1))

    # 有新老两种 symbol 时对传入的 underlying_symbols 需要对应成新的 symbol, 并对并行期结束后仍使用老的 symbol 予以警告
    multi_symbol_map = {'RO': 'OI', 'WS': 'WH', 'ER': 'RI', 'TC': 'ZC', 'ME': 'MA'}
    symbol_date_map = {'RO': 20130515, 'WS': 20130523, 'ER': 20130523, 'TC': 20160408, 'ME': 20150515}
    for symbol in set(underlying_symbols) & set(multi_symbol_map):
        date_line = symbol_date_map[symbol]
        if end_date > date_line:
            import warnings
            msg = 'You are using the old symbol: {}, however the new symbol: {} is available after {}.'.format(symbol, multi_symbol_map[symbol], date_line)
            warnings.warn(msg, category=DeprecationWarning, stacklevel=2)

    # 对传入的 underlying_symbols 依照 multi_symbol_map 生成一个对照 DataFrame
    symbol_map_df = pd.DataFrame([(symbol, multi_symbol_map.get(symbol, symbol)) for symbol in set(underlying_symbols)],
                                 columns=['origin', 'new'])
    # 将 underlying_symbols 中 所有老的 symbol 对应为新的再去 mongo 查询
    underlying_symbols = list(symbol_map_df.new.unique())
    ret = get_client().execute("futures.get_warehouse_stocks", underlying_symbols, start_date, end_date, market=market)
    if not ret:
        return
    columns = ["date", "underlying_symbol", "on_warrant", "exchange", 'effective_forecast', 'warrant_units',
               'contract_multiplier', 'deliverable']
    df = pd.DataFrame(ret, columns=columns)

    df = df.merge(symbol_map_df, left_on='underlying_symbol', right_on='new')
    df.drop(['underlying_symbol', 'new'], axis=1, inplace=True)
    df.rename(columns={'origin': 'underlying_symbol'}, inplace=True)
    df.set_index(['date', 'underlying_symbol'], inplace=True)
    return df.sort_index()


@export_as_api(namespace="futures")
def get_contract_multiplier(underlying_symbols, start_date=None, end_date=None, market="cn"):
    """获取时间区间内期货的合约乘数

    :param underlying_symbols: 期货品种, 支持列表查询
    :param start_date: 开始日期, 如'2015-01-01', 如果不填写则取underlying_symbols对应实际数据最早范围
    :param end_date: 结束日期, 如'2015-01-01', 如果不填写则为当日前一天
    :param market: 市场, 默认为"cn", 当前仅支持中国市场
    :return: pd.DataFrame

    """
    underlying_symbols = ensure_list_of_string(underlying_symbols, name="underlying_symbols")
    ret = get_client().execute("futures.get_contract_multiplier", underlying_symbols)
    if not ret:
        return

    # 因 mongo 数据为时间范围，要返回每一天的数据，需复制合约乘数数据至至范围内所有 trading_date
    if start_date:
        start_date = to_datetime(start_date)
    if not end_date:
        end_date = datetime.datetime.today() - datetime.timedelta(days=1)
    end_date = to_datetime(end_date)

    def fill(group_df):
        # 根据当前合约日期范围及给定范围内获取所有 trading_date
        date_min, date_max = group_df['effective_date'].min(), group_df['cancel_date'].max()
        if start_date is not None:
            date_min = max(start_date, date_min)
        date_max = min(date_max, end_date)
        trading_dates = pd.to_datetime(get_trading_dates(date_min, date_max)+group_df['effective_date'].to_list()).unique()

        # 使用 trading_dates 作为 index 插入并填充数据
        everyday_df = group_df.set_index(['effective_date']).reindex(trading_dates).sort_index().ffill().reset_index().rename(columns={'index': 'date'})
        everyday_df = everyday_df[(everyday_df['date'] >= date_min) & (everyday_df['date'] <= date_max)]

        return everyday_df

    df = pd.DataFrame(ret).groupby(by=['underlying_symbol']).apply(fill)

    df = df[['date', 'underlying_symbol', 'exchange', 'contract_multiplier']]
    df.set_index(['underlying_symbol', 'date'], inplace=True)
    df.sort_index(inplace=True)
    return df


@export_as_api(namespace="futures")
def get_basis(order_book_ids, start_date=None, end_date=None, fields=None, market="cn"):
    """ 获取股指期货升贴水信息.

    :param order_book_ids: 期货合约, 支持 order_book_id 或 order_book_id list.
    :param start_date: 开始时间, 若不传, 为 end_date 前3个月.
    :param end_date: 结束时间, 若不传, 为 start_date 后3个月, 如果 start_date 也不传, 则默认为最近3个月.
    :param fields: 需要返回的字段, 若不传则返回所有字段, 支持返回的字段包括
        open, high, low, close, index, close_index, basis, basis_rate, basis_annual_rate.
    :return: MultiIndex DataFrame.
    """
    start_date, end_date = ensure_date_range(start_date, end_date)
    def _calc_annual_rate(row):
        # row.name[1] is current date.
        order_book_id, current_date = row.name
        dates_to_delisted = calendar.get_trading_dates(current_date, delisted_map[order_book_id])[1:]
        if len(dates_to_delisted) == 0:
            # 在到期的时候, basis_annual_rate 的值本身也没有什么意义, 所以直接赋值为 nan.
            return float("nan")
        else:
            return row["basis_rate"] * 250 / len(dates_to_delisted)

    # filter and check order_book_ids.
    from ..services import basic, get_price, calendar
    from ..services.detail import get_price_df
    insts = basic.instruments(order_book_ids)
    if insts is None:
        return None
    if not isinstance(order_book_ids, list):
        insts = [insts]
    insts = [
        x for x in insts
        if x.type == "Future" and x.listed_date != "0000-00-00" and x.industry_name == "股指"
    ]
    if not insts:
        return None
    underlying_id_map = {x.order_book_id: x.underlying_order_book_id for x in insts}

    # get future price.
    order_book_ids = [x.order_book_id for x in insts]
    future_price = get_price.get_price(order_book_ids, start_date, end_date, fields=["close", "open", "high", "low"], expect_df=True)
    if future_price is None:
        return None
    future_price["index"] = future_price.index.get_level_values("order_book_id").map(underlying_id_map)
    # get index price.
    underlying_ids = list({x.underlying_order_book_id for x in insts})
    index_close = get_price_df.get_future_indx_daybar(underlying_ids, start_date, end_date, fields=["close"])["close"]

    # x.name[1] is relative timestamp.
    future_price["close_index"] = future_price.apply(lambda x: index_close.loc[(x['index'], x.name[1])], axis=1)
    future_price["basis"] = future_price["close"] - future_price["close_index"]
    future_price["basis_rate"] = future_price["basis"] / future_price["close_index"] * 100
    delisted_map = {x.order_book_id: x.de_listed_date for x in insts}
    future_price["basis_annual_rate"] = future_price.apply(lambda x: _calc_annual_rate(x), axis=1)

    # filter by given fields.
    VALID_FIELDS = {"open", "high", "low", "close", "index", "close_index", "basis", "basis_rate", "basis_annual_rate"}
    if fields is not None:
        fields = set(ensure_list_of_string(fields))
    else:
        fields = VALID_FIELDS

    fields = list(fields & VALID_FIELDS)
    return future_price[fields]


VALID_ADJUST_METHODS = ['prev_close_spread', 'open_spread', 'prev_close_ratio', 'open_ratio']


@ttl_cache(1800)
def _get_future_factors_df(market='cn'):
    """ 获取所有复权因子表 """
    data = get_client().execute('futures.__internal__get_future_factors', market=market)
    df = pd.DataFrame(data)
    df['ex_date'] = df['ex_date'].apply(int8_to_datetime)
    df.set_index(['underlying_symbol', 'ex_date'], inplace=True)
    df.sort_index(inplace=True)
    return df


@export_as_api(namespace='futures')
def get_ex_factor(underlying_symbols, start_date=None, end_date=None, adjust_method='prev_close_spread', market='cn'):
    """ 获取期货复权因子

    :param underlying_symbols: 期货合约品种，str or list
    :param start_date: 开始日期
    :param end_date: 结束日期
    :param adjust_method: 复权方法，prev_close_spread, prev_close_ratio, open_spread, open_ratio,
    默认为‘prev_close_spread'
    :param market: 默认是中国内地市场('cn') 。可选'cn' - 中国内地市场
    :return: DataFrame
    """
    df = _get_future_factors_df(market)
    valid_underlying_symbols = df.index.get_level_values('underlying_symbol').unique().tolist()
    underlying_symbols = ensure_list_of_string(underlying_symbols, 'underlying_symbols')
    check_items_in_container(adjust_method, VALID_ADJUST_METHODS, 'adjust_method')
    check_items_in_container(underlying_symbols, valid_underlying_symbols, 'underlying_symbols')

    factor = df.loc[underlying_symbols, adjust_method]
    factor.name = 'ex_factor'
    factor = factor.reset_index()

    spread = adjust_method.endswith('spread')

    def _process(x):
        x['ex_end_date'] = x['ex_date'].shift(-1) - pd.offsets.DateOffset(days=1)
        if spread:
            x['ex_cum_factor'] = x['ex_factor'].cumsum()
        else:
            x['ex_cum_factor'] = x['ex_factor'].cumprod()
        return x

    factor = factor.groupby('underlying_symbol', as_index=False).apply(_process)
    if start_date and end_date:
        start_date, end_date = to_datetime(start_date), to_datetime(end_date)
        factor = factor[(start_date <= factor['ex_date']) & (factor['ex_date'] <= end_date)]
    # _get_future_factors_df 已经排序过了，此处无需再次排序
    return factor.set_index('ex_date')


@ttl_cache(1800)
def __internal_get_ex_factor(adjust_type, adjust_method):
    """ 内部使用，获取复权因子，提供给get_dominant_price进行复权计算用
    :return: pd.Series
    """
    df = _get_future_factors_df()

    factor = df[adjust_method]
    factor.name = 'ex_factor'
    factor = factor.reset_index()
    pre = adjust_type == 'pre'
    ratio = adjust_method.endswith('ratio')
    today = datetime.date.today()

    def _process(x):
        if ratio:
            x['ex_cum_factor'] = x['ex_factor'].cumprod()
            if pre:
                x['ex_cum_factor'] = x['ex_cum_factor'] / x['ex_cum_factor'].iloc[-1]
        else:
            x['ex_cum_factor'] = x['ex_factor'].cumsum()
            if pre:
                x['ex_cum_factor'] = x['ex_cum_factor'] - x['ex_cum_factor'].iloc[-1]

        # 按交易日扩展后ffill
        min_date = x['ex_date'].min()
        tds = get_trading_dates(min_date, today)
        tds = pd.to_datetime(tds)
        # tds 是从小到大排列的， 因此reindex后无需再sort
        return x.set_index('ex_date').reindex(tds).ffill()

    factor = factor.groupby('underlying_symbol', as_index=True).apply(_process)
    return factor['ex_cum_factor']


DOMINANT_PRICE_ADJUST_FIELDS = [
    'open', 'high', 'low', 'close', 'last', 'limit_up', 'limit_down', 'settlement', 'prev_settlement', 'prev_close',
    'a1', 'a2', 'a3', 'a4', 'a5', 'b1', 'b2', 'b3', 'b4', 'b5'
]


@export_as_api(namespace='futures')
def get_dominant_price(
        underlying_symbols, start_date=None, end_date=None,
        frequency='1d', fields=None, adjust_type='pre', adjust_method='prev_close_spread'
):
    """ 获取主力合约行情数据

    :param underlying_symbols: 期货合约品种，可传入 underlying_symbol, underlying_symbol list
    :param start_date: 开始日期, 最小日期为 20210104
    :param end_date: 结束日期
    :param frequency: 历史数据的频率。 支持/日/分钟/tick 级别的历史数据，默认为'1d'。
        1m- 分钟线，1d-日线，分钟可选取不同频率，例如'5m'代表 5 分钟线
    :param fields: 字段名称列表
    :param adjust_type: 复权方式，不复权 - none，前复权 - pre，后复权 - post
    :param adjust_method: 复权方法 ，prev_close_spread/open_spread:基于价差复权因子进行复权，
        prev_close_ratio/open_ratio:基于比例复权因子进行复权，
        默认为‘prev_close_spread',adjust_type为None 时，adjust_method 复权方法设置无效
    :return: MultiIndex DataFrame
    """
    # ensure underlying symbols list
    from ..services.get_price import get_price
    if not isinstance(underlying_symbols, list):
        underlying_symbols = [underlying_symbols]
    if fields and not isinstance(fields, list):
        fields = [fields]
    start_date, end_date = ensure_date_range(start_date, end_date)
    if start_date < 20100104:
        raise ValueError('expect start_date >= 20100104, get {}'.format(start_date))
    # ensure adjust_type and adjust_method
    check_items_in_container(adjust_type, ['none', 'pre', 'post'], 'adjust_type')
    check_items_in_container(adjust_method, VALID_ADJUST_METHODS, 'adjust_method')

    _date_key = 'date' if frequency == '1d' else 'trading_date'
    _fields = fields
    if fields and frequency != '1d' and 'trading_date' not in fields:
        _fields = ['trading_date'] + fields

    obs = [u + '88' for u in underlying_symbols]
    df = get_price(
        order_book_ids=obs, start_date=start_date, end_date=end_date,
        frequency=frequency, adjust_type='none', fields=_fields, expect_df=True
    )
    if df is None:
        return

    df.reset_index(inplace=True)
    df['underlying_symbol'] = df['order_book_id'].str[:-2]
    df.set_index(['underlying_symbol', _date_key], inplace=True)
    if adjust_type != 'none':
        # 复权调整
        factor = __internal_get_ex_factor(adjust_type, adjust_method)
        values = factor.reindex(df.index).values
        _fields = fields if fields else df.columns.tolist()
        adjust_fields = [f for f in DOMINANT_PRICE_ADJUST_FIELDS if f in _fields]
        if adjust_method.endswith('spread'):
            for field in adjust_fields:
                df[field] += values
        elif adjust_method.endswith('ratio'):
            for field in adjust_fields:
                df[field] *= values
        if 'total_turnover' in df.columns:
            df['total_turnover'] = 0

    if frequency != '1d':
        df = df.reset_index().set_index(['underlying_symbol', 'datetime'])
    df.sort_index(inplace=True)
    del df['order_book_id']
    return df[fields] if fields else df



# coding: utf-8
from sqlalchemy import Numeric, Column
from .fundamental_base_sql import FundamentalBase


class StkIncomeGen(FundamentalBase):
    revenue = Column(Numeric(18, 2))
    operating_revenue = Column(Numeric(18, 2))
    sales_discount = Column(Numeric(18, 2))
    net_operating_revenue = Column(Numeric(18, 2))
    total_expense = Column(Numeric(18, 2))
    cost_of_goods_sold = Column(Numeric(18, 2))
    sales_tax = Column(Numeric(18, 2))
    gross_profit = Column(Numeric(18, 2))
    other_operating_income = Column(Numeric(18, 2))
    inventory_shrinkage = Column(Numeric(18, 2))
    selling_expense = Column(Numeric(18, 2))
    operating_expense = Column(Numeric(18, 2))
    ga_expense = Column(Numeric(18, 2))
    financing_expense = Column(Numeric(18, 2))
    exchange_gains_or_losses = Column(Numeric(18, 2))
    asset_depreciation = Column(Numeric(18, 2))
    profit_from_operation = Column(Numeric(18, 2))
    investment_income = Column(Numeric(18, 2))
    subsidy_income = Column(Numeric(18, 2))
    non_operating_revenue = Column(Numeric(18, 2))
    pnl_adjustment = Column(Numeric(18, 2))
    non_operating_expense = Column(Numeric(18, 2))
    disposal_loss_on_asset = Column(Numeric(18, 2))
    non_operating_net_profit = Column(Numeric(18, 2))
    profit_before_tax = Column(Numeric(18, 2))
    income_tax = Column(Numeric(18, 2))
    profit_from_ma = Column(Numeric(18, 2))
    net_profit = Column(Numeric(18, 2))
    net_profit_parent_company = Column(Numeric(18, 2))
    retained_profit_at_beginning = Column(Numeric(18, 2))
    profit_available_for_distribution = Column(Numeric(18, 2))
    statutory_welfare_reserve = Column(Numeric(18, 2))
    staff_incentive_welfare_reserve = Column(Numeric(18, 2))
    enterprise_expansion_reserve = Column(Numeric(18, 2))
    profit_available_for_owner_distribution = Column(Numeric(18, 2))
    preferred_stock_dividends = Column(Numeric(18, 2))
    other_surplus_reserve = Column(Numeric(18, 2))
    ordinary_stock_dividends = Column(Numeric(18, 2))
    basic_earnings_per_share = Column(Numeric(18, 4))
    other_income = Column(Numeric(18, 2))
    total_income = Column(Numeric(18, 2))
    total_income_parent_company = Column(Numeric(18, 2))
    total_income_minority = Column(Numeric(18, 2))

    fair_value_change_income = Column(Numeric(18, 2))
    invest_income_associates = Column(Numeric(18, 2))
    interest_income = Column(Numeric(18, 2))
    interest_expense = Column(Numeric(18, 2))

    other_operating_revenue = Column(Numeric(19, 4))
    other_operating_cost = Column(Numeric(19, 4))
    asset_impairment = Column(Numeric(19, 4))

    r_n_d = Column(Numeric(19, 4))
    minority_profit = Column(Numeric(19, 4))
    non_recurring_pnl = Column(Numeric(19, 4))
    net_profit_deduct_non_recurring_pnl = Column(Numeric(19, 4))



# -*- coding: utf-8 -*-
#
# Copyright 2018 Ricequant, Inc
import os
import logging

from . import client


class _RQDataClient:
    def __init__(self):
        import rqdatad.core.app
        import rqdatad.__main__
        import rqdatad.database

        app = rqdatad.core.app.app
        app.config = rqdatad.__main__.default_config_file()
        app.init()
        rqdatad.database.set_database_config(app.config)

        self.app = app
        self.PID = os.getpid()

    def execute(self, api_name, *args, **kwargs):
        return self.app.debug_api(api_name, *args, **kwargs)

    def close(self):
        import asyncio

        loop = self.app.get_event_loop()
        loop.stop()
        for task in asyncio.Task.all_tasks():
            task.cancel()
        loop.close()


def as_rqdata():
    logging.info('RQDATAC: behave like RQDATA now')
    client._CLIENT = _RQDataClient()



# -*- coding: utf-8 -*-
import threading
import socket

from rqdatac.share.errors import ErrorFromBackend, GatewayError
from rqdatac.connection import Connection
from rqdatac.decorators import retry
from rqdatac.utils import connection_error, timeout_error


class ThreadLocalConnection:
    def __init__(self, addr, auth=None, connect_timeout=5, timeout=60):
        self._local = threading.local()
        self._addr = addr
        self._auth = auth
        self._connect_timeout = connect_timeout
        self._timeout = timeout

    def _get_connection(self):
        conn = getattr(self._local, "connection", None)
        if conn is not None:
            if conn.is_normal():
                return conn
            else:
                self.close()
        s = Connection.sock_factory(self._addr, timeout=self._connect_timeout)
        s.settimeout(self._timeout)
        self._local.connection = Connection(s, self._auth)
        return self._local.connection

    def _execute(self, conn, method, args, kwargs):
        try:
            return conn.execute(method, *args, **kwargs)
        except (KeyboardInterrupt, Exception) as e:
            if not isinstance(e, (ErrorFromBackend, GatewayError)):
                self.reset()
                conn.close()
            raise e

    @retry(3, suppress_exceptions=(connection_error, GatewayError, timeout_error, socket.timeout))
    def execute(self, method, *args, **kwargs):
        return self._execute(self._get_connection(), method, args, kwargs)

    @retry(3, suppress_exceptions=(connection_error, GatewayError))
    def execute_with_timeout(self, timeout, method, *args, **kwargs):
        assert isinstance(timeout, int) and timeout > 0
        conn = self._get_connection()
        conn.set_timeout(timeout)
        r = self._execute(conn, method, args, kwargs)
        conn.set_timeout(self._timeout)
        return r

    def close(self):
        if getattr(self._local, "connection", None) is not None:
            self._local.connection.close()
            self._local.connection = None

    reset = close



# coding: utf-8
from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
metadata = Base.metadata


class FundamentalBase(Base):
    __tablename__ = "fundamental_view"

    stockcode = Column(String(12), primary_key=True, nullable=False)
    tradedate = Column(Integer, primary_key=True, nullable=False)
    end_date = Column(Integer)
    announce_date = Column(Integer)
    rpt_year = Column(Integer)
    rpt_quarter = Column(Integer)



# -*- coding: utf-8 -*-
import datetime
from itertools import chain

import pandas as pd
import numpy as np

from rqdatac.services.basic import instruments
from rqdatac.services.future import get_dominant_future
from rqdatac.services.calendar import get_trading_dates_in_type, get_previous_trading_date, is_trading_date, get_next_trading_date
from rqdatac.utils import (to_datetime, int8_to_date, int9_to_time, int14_to_datetime,
                           get_tick_value, datetime_to_int17, relativedelta, int17_to_datetime)
from rqdatac.client import get_client
from rqdatac.decorators import export_as_api, ttl_cache
from rqdatac.validators import (ensure_instruments_in_order,
                                check_items_in_container, ensure_list_of_string)
from rqdatac.services.stock_status import is_suspended


@ttl_cache(3600)
def _to_real_contract(ob, market):
    today = datetime.date.today()
    r = get_dominant_future(ob, today, today, market=market)
    if isinstance(r, pd.Series) and r.size == 1:
        return r[0]
    return None


@export_as_api
def current_snapshot(order_book_ids, market="cn"):
    """获取实时行情

    :param order_book_ids: 股票或期货代码或代码列表
    :param market: 地区代码, 如'cn' (Default value = "cn")
    :returns: Snapshot Object或Snapshot Object列表

    """
    _instruments = ensure_instruments_in_order(order_book_ids)
    order_book_ids = []
    for ins in _instruments:
        if ins.type == "Future" and ins.order_book_id.endswith(("88", "89")):
            real_contract = _to_real_contract(ins.underlying_symbol, market) or ins.order_book_id
            order_book_ids.append(real_contract)
        else:
            order_book_ids.append(ins.order_book_id)

    snapshots = get_client().execute("current_snapshots", order_book_ids, market=market)
    tick_objects = [TickObject(ins.order_book_id, snapshots[i]) for i, ins in enumerate(_instruments)]
    if len(order_book_ids) == 1:
        return tick_objects[0]
    return tick_objects


class TickObject(object):
    _STOCK_FIELDS = [
        ("datetime", np.uint64),
        ("open", np.float64),
        ("high", np.float64),
        ("low", np.float64),
        ("last", np.float64),
        ('limit_up', np.float64),
        ('limit_down', np.float64),
        ("iopv", np.float64),
        ("pre_iopv", np.float64),
        ("volume", np.int32),
        ("total_turnover", np.int64),
        ("prev_close", np.float64),
        ("ask", list),
        ("ask_vol", list),
        ("bid", list),
        ("bid_vol", list),
        ("trading_phase_code", str),
    ]

    _FUTURE_FIELDS = _STOCK_FIELDS + [("open_interest", np.int32), ("prev_settlement", np.float64)]

    _STOCK_FIELD_NAMES = [_n for _n, _ in _STOCK_FIELDS]
    _FUTURE_FIELD_NAMES = [_n for _n, _ in _FUTURE_FIELDS]

    _NANDict = {_n: np.nan for _n in _FUTURE_FIELD_NAMES}

    def __init__(self, order_book_id, data, dt=None):
        self._dt = dt
        if data is None:
            self._data = self._NANDict
        else:
            self._data = data
        self._order_book_id = order_book_id

    @property
    def order_book_id(self):
        return self._order_book_id

    @property
    def open(self):
        return self._data.get("open", 0)

    @property
    def last(self):
        return self._data.get("last", 0)

    @property
    def low(self):
        return self._data.get("low", 0)

    @property
    def high(self):
        return self._data.get("high", 0)

    @property
    def limit_up(self):
        return self._data.get('limit_up', 0)

    @property
    def limit_down(self):
        return self._data.get('limit_down', 0)

    @property
    def num_trades(self):
        return self._data.get('num_trades', 0)

    @property
    def prev_close(self):
        return self._data.get("prev_close", 0)

    @property
    def iopv(self):
        return self._data.get("iopv", np.nan)

    @property
    def prev_iopv(self):
        return self._data.get("prev_iopv", np.nan)

    @property
    def volume(self):
        return self._data.get("volume", 0)

    @property
    def total_turnover(self):
        return self._data.get("total_turnover", 0)

    @property
    def datetime(self):
        if self._dt is not None:
            return self._dt
        if not self._isnan:
            dt = self._data["datetime"]
            return to_datetime(dt)
        return datetime.datetime.min

    @property
    def prev_settlement(self):
        try:
            return self._data["prev_settlement"]
        except KeyError:
            return None

    @property
    def open_interest(self):
        try:
            return self._data["open_interest"]
        except KeyError:
            return None

    @property
    def trading_phase_code(self):
        try:
            return self._data["trading_phase_code"]
        except KeyError:
            return None

    @property
    def asks(self):
        return self._data.get("ask", [0, 0, 0, 0, 0])

    @property
    def ask_vols(self):
        return self._data.get("ask_vol", [0, 0, 0, 0, 0])

    @property
    def bids(self):
        return self._data.get("bid", [0, 0, 0, 0, 0])

    @property
    def bid_vols(self):
        return self._data.get("bid_vol", [0, 0, 0, 0, 0])

    @property
    def _isnan(self):
        return np.isnan(self._data["last"])

    def __repr__(self):
        items = []
        for name in dir(self):
            if name.startswith("_"):
                continue
            items.append((name, getattr(self, name)))
        return "Tick({0})".format(
            ", ".join("{0}: {1}".format(k, v) for k, v in items if k is not None)
        )

    def __getitem__(self, key):
        return getattr(self, key)


EQUITIES_FIELDS = (
    "datetime",
    "open",
    "last",
    "high",
    "low",
    "iopv",
    "prev_iopv",
    "limit_up",
    "limit_down",
    "prev_close",
    "volume",
    "total_turnover",
    "a1",
    "a2",
    "a3",
    "a4",
    "a5",
    "b1",
    "b2",
    "b3",
    "b4",
    "b5",
    "a1_v",
    "a2_v",
    "a3_v",
    "a4_v",
    "a5_v",
    "b1_v",
    "b2_v",
    "b3_v",
    "b4_v",
    "b5_v",
)

FUTURE_FIELDS = (
    "datetime",
    "trading_date",
    "update_time",
    "open",
    "last",
    "high",
    "low",
    "limit_up",
    "limit_down",
    "prev_settlement",
    "prev_close",
    "volume",
    "total_turnover",
    "open_interest",
    "a1",
    "a2",
    "a3",
    "a4",
    "a5",
    "b1",
    "b2",
    "b3",
    "b4",
    "b5",
    "a1_v",
    "a2_v",
    "a3_v",
    "a4_v",
    "a5_v",
    "b1_v",
    "b2_v",
    "b3_v",
    "b4_v",
    "b5_v",
)


@export_as_api
def get_ticks(order_book_id, start_date=None, end_date=None, expect_df=True, market="cn"):
    """获取实时tick

    :param order_book_id: 股票代码, 如'000001.XSHE'
    :param start_date: 开始日期, 不指定则为今天
    :param end_date: 结束日期, 不指定则为今天
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :param market:  (Default value = "cn")
    :returns: pd.DataFrame or None

    """
    instrument = instruments(order_book_id, market)
    if not instrument:
        raise ValueError("invalid order_book_id: {}".format(order_book_id))

    order_book_id = instrument.order_book_id
    future_like = instrument.type in ("Future", "Option", "Spot")

    now = datetime.datetime.now()
    if is_trading_date(now):
        if (now.hour, now.minute) >= (19, 30):
            day = get_next_trading_date(now)
        else:
            day = now.date()
    else:
        day = get_next_trading_date(now)

    dates = get_trading_dates_in_type(
        start_date or day,
        end_date or day,
        expect_type="str",
        fmt="%Y%m%d",
        market=market,
    )
    ticks = get_client().execute("get_ticks_v2", order_book_id, dates, market=market)
    if future_like:
        fields = FUTURE_FIELDS
    else:
        fields = EQUITIES_FIELDS

    if future_like and instrument.exchange in ("XSHE", "XSHG"):
        fields = list(fields)
        fields.remove("trading_date")

    dtype = np.dtype([(f, _field_type(f)) for f in fields])
    bars = np.array([tuple([get_tick_value(t, f, np.nan) if f in ("iopv", "prev_iopv") else get_tick_value(t, f)
                            for f in fields]) for t in chain(*ticks)], dtype=dtype)
    df = pd.DataFrame(bars)

    if df.empty:
        return None
    if "trading_date" in df.columns:
        df.trading_date = df.trading_date.apply(int8_to_date)
        df.update_time = df.update_time.apply(int9_to_time)
    df.datetime = df.datetime.apply(to_datetime)
    if expect_df:
        df["order_book_id"] = order_book_id
        df.set_index(["order_book_id", "datetime"], inplace=True)
    else:
        df.set_index("datetime", inplace=True)
    return df


def _field_type(field_name):
    return np.uint64 if field_name in ("datetime", "trading_date", "update_time") else np.float


@export_as_api
def current_minute(order_book_ids, skip_suspended=False, fields=None, market="cn"):
    """获取实时分钟行情
    :param order_book_ids: 合约代码或代码列表
    :param skip_suspended: 是否过滤停复牌. (Default value = False)
    :param fields: 可选参数。默认为所有字段。 (Default value = None)
    :param market: 地区代码, 如'cn' (Default value = "cn")
    :returns: None or pd.DataFrame
    """
    from rqdatac.services.get_price import classify_order_book_ids, _ensure_fields
    from rqdatac.services.detail.get_price_df import MINBAR_FIELDS

    (order_book_ids, stocks, funds, indexes, futures, futures888, spots, options,
        convertibles, repos) = classify_order_book_ids(order_book_ids, market)
    if skip_suspended and stocks:
        date = get_previous_trading_date(datetime.date.today() + datetime.timedelta(days=1))
        df_suspended = is_suspended(stocks, date, date)
        if not df_suspended.empty:
            df_suspended_t = df_suspended.T
            suspended_obids = set(df_suspended_t[df_suspended_t[df_suspended_t.columns[0]]].index)
            inspection = suspended_obids & set(stocks)
            if inspection:
                stocks = set(stocks) - inspection
                order_book_ids = list(set(order_book_ids) - inspection)

    fields, _ = _ensure_fields(fields, MINBAR_FIELDS, stocks, funds, futures, futures888, spots, options, convertibles, indexes, repos)

    data = get_client().execute("current_minute", order_book_ids, fields + ["datetime"], market=market)
    if not data:
        return
    df = pd.DataFrame(data)
    df["datetime"] = df["datetime"].map(int14_to_datetime, na_action="ignore")
    df.set_index(["order_book_id", "datetime"], inplace=True)
    return df


LIVE_EQUITIES_FIELDS = [
    "open",
    "last",
    "high",
    "low",
    "iopv",
    "prev_iopv",
    "limit_up",
    "limit_down",
    "prev_close",
    "volume",
    "total_turnover",
    "a1",
    "a2",
    "a3",
    "a4",
    "a5",
    "b1",
    "b2",
    "b3",
    "b4",
    "b5",
    "a1_v",
    "a2_v",
    "a3_v",
    "a4_v",
    "a5_v",
    "b1_v",
    "b2_v",
    "b3_v",
    "b4_v",
    "b5_v",
    "num_trades",
]

LIVE_FUTURE_FIELDS = [
    "trading_date",
    "update_time",
    "open",
    "last",
    "high",
    "low",
    "limit_up",
    "limit_down",
    "prev_settlement",
    "prev_close",
    "volume",
    "total_turnover",
    "open_interest",
    "a1",
    "a2",
    "a3",
    "a4",
    "a5",
    "b1",
    "b2",
    "b3",
    "b4",
    "b5",
    "a1_v",
    "a2_v",
    "a3_v",
    "a4_v",
    "a5_v",
    "b1_v",
    "b2_v",
    "b3_v",
    "b4_v",
    "b5_v",
]


@export_as_api
def get_live_ticks(order_book_ids, start_dt=None, end_dt=None, fields=None, market="cn"):
    """获取实时tick
    :param order_book_ids: 股票代码或代码列表, 如'000001.XSHE'
    :param start_dt: 开始时间，如20200102213000 或 "2020-01-02 21:30:00"， 默认为今天
    :param end_dt: 结束时间，如20200102153000 或 "2020-01-03 15:30:00"， 默认为今天
    :param fields: 可选参数。默认为所有字段。 (Default value = None)
    :param market:  (Default value = "cn")

    :returns: pd.DataFrame or None
    """

    if start_dt is None and end_dt is None:
        now = datetime.datetime.now()
        if is_trading_date(now):
            if (now.hour, now.minute) >= (19, 30):
                day = get_next_trading_date(now)
            else:
                day = now.date()
        else:
            day = get_next_trading_date(now)

        dates = get_trading_dates_in_type(
            day, day,
            expect_type="str",
            fmt="%Y%m%d",
            market=market,
        )
        start_dt, end_dt = 0, 29991231240000000
    elif start_dt and end_dt:
        start_datetime = to_datetime(start_dt)
        end_datetime = to_datetime(end_dt)
        start_dt = datetime_to_int17(start_datetime)
        end_dt = datetime_to_int17(end_datetime)

        dates = get_trading_dates_in_type(
            start_datetime + relativedelta(days=1 if start_datetime.hour > 18 else 0),
            end_datetime + relativedelta(days=1 if start_datetime.hour > 18 else 0),
            expect_type="str",
            fmt="%Y%m%d",
            market=market,
        )
    else:
        raise ValueError("please specify start_dt/end_dt in the same time")
    order_book_ids = ensure_list_of_string(order_book_ids)
    ins = instruments(order_book_ids, market)
    if not ins:
        raise ValueError("invalid order_book_id: {}".format(order_book_ids))
    obids = [i.order_book_id for i in ins]
    ins_types = {i.type for i in ins}

    future_like = ins_types & {"Future", "Option", "Spot"}
    equities_like = ins_types - {"Future", "Option", "Spot"}
    if future_like and equities_like:
        base_fields = LIVE_EQUITIES_FIELDS + list(set(LIVE_FUTURE_FIELDS) - set(LIVE_EQUITIES_FIELDS))
    elif future_like:
        base_fields = LIVE_FUTURE_FIELDS
    else:
        base_fields = LIVE_EQUITIES_FIELDS

    if fields is not None:
        fields = ensure_list_of_string(fields)
        check_items_in_container(fields, base_fields, "fields")
    else:
        fields = base_fields
    fields = ["order_book_id", "datetime"] + fields
    ret = get_client().execute("get_live_ticks", obids, start_dt, end_dt, dates, fields, market=market)
    if not ret:
        return None
    header, ticks = ret
    df = pd.DataFrame(ticks, columns=header)
    df = df[fields]
    if "trading_date" in df.columns:
        df.trading_date = df.trading_date.apply(lambda x: pd.NaT if x == 0 else int8_to_date(x))
    if "update_time" in df.columns:
        df.update_time = df.update_time.apply(int9_to_time)
    df["datetime"] = df["datetime"].apply(int17_to_datetime)
    df.set_index(["order_book_id", "datetime"], inplace=True)
    df.sort_index(inplace=True)
    return df



# coding: utf-8
class PitBase:
    pass


class BalanceSheet(PitBase):
    deferred_expense = "deferred_expense"
    dividend_payable = "dividend_payable"
    short_term_debt = "short_term_debt"
    financial_liabilities = "financial_liabilities"
    equity_parent_company = "equity_parent_company"
    tax_payable = "tax_payable"
    payroll_payable = "payroll_payable"
    intangible_assets = "intangible_assets"
    deferred_revenue = "deferred_revenue"
    total_assets = "total_assets"
    depreciation_reserve = "depreciation_reserve"
    accts_payable = "accts_payable"
    deferred_income_tax_liabilities = "deferred_income_tax_liabilities"
    other_fees_payable = "other_fees_payable"
    financial_asset_hold_to_maturity = "financial_asset_hold_to_maturity"
    construction_in_progress = "construction_in_progress"
    goodwill = "goodwill"
    other_payable = "other_payable"
    long_term_receivables = "long_term_receivables"
    accrued_expense = "accrued_expense"
    capital_reserve = "capital_reserve"
    accumulated_depreciation = "accumulated_depreciation"
    surplus_reserve = "surplus_reserve"
    real_estate_investment = "real_estate_investment"
    engineer_material = "engineer_material"
    minority_interest = "minority_interest"
    provision = "provision"
    oil_and_gas_assets = "oil_and_gas_assets"
    long_term_liabilities_due_one_year = "long_term_liabilities_due_one_year"
    bond_payable = "bond_payable"
    net_fixed_assets = "net_fixed_assets"
    cash_equivalent = "cash_equivalent"
    financial_asset_available_for_sale = "financial_asset_available_for_sale"
    estimated_liabilities = "estimated_liabilities"
    bad_debt_reserve = "bad_debt_reserve"
    other_non_current_liabilities = "other_non_current_liabilities"
    non_current_liability_due_one_year = "non_current_liability_due_one_year"
    non_current_assets = "non_current_assets"
    long_term_deferred_expenses = "long_term_deferred_expenses"
    net_long_term_equity_investment = "net_long_term_equity_investment"
    bill_receivable = "bill_receivable"
    non_current_asset_due_one_year = "non_current_asset_due_one_year"
    total_equity_and_liabilities = "total_equity_and_liabilities"
    long_term_equity_investment = "long_term_equity_investment"
    total_liabilities = "total_liabilities"
    prepaid_tax = "prepaid_tax"
    grants_received = "grants_received"
    paid_in_capital = "paid_in_capital"
    housing_revolving_funds = "housing_revolving_funds"
    notes_payable = "notes_payable"
    long_term_payable = "long_term_payable"
    other_non_current_assets = "other_non_current_assets"
    interest_receivable = "interest_receivable"
    impairment_intangible_assets = "impairment_intangible_assets"
    advance_from_customers = "advance_from_customers"
    deferred_income_tax_assets = "deferred_income_tax_assets"
    other_current_assets = "other_current_assets"
    fixed_asset_to_be_disposed = "fixed_asset_to_be_disposed"
    financial_asset_held_for_trading = "financial_asset_held_for_trading"
    short_term_loans = "short_term_loans"
    inventory = "inventory"
    deferred_income = "deferred_income"
    interest_payable = "interest_payable"
    non_current_liabilities = "non_current_liabilities"
    dividend_receivable = "dividend_receivable"
    other_current_liabilities = "other_current_liabilities"
    current_assets = "current_assets"
    long_term_loans = "long_term_loans"
    capitalized_biological_assets = "capitalized_biological_assets"
    current_liabilities = "current_liabilities"
    contract_work = "contract_work"
    total_equity = "total_equity"
    prepayment = "prepayment"
    net_accts_receivable = "net_accts_receivable"
    undistributed_profit = "undistributed_profit"
    equity_preferred_stock = "equity_preferred_stock"
    other_accts_receivable = "other_accts_receivable"
    other_equity_investment = "other_equity_investment"
    other_illiquidy_financial_assets = "other_illiquidy_financial_assets"
    other_payable_interest_dividend = "other_payable_interest_dividend"


class CashFlowsStatement(PitBase):
    exchange_rate_change_effect = "exchange_rate_change_effect"
    cash_paid_for_other_financing_activities = "cash_paid_for_other_financing_activities"
    cash_received_from_investors = "cash_received_from_investors"
    cash_from_operating_activities = "cash_from_operating_activities"
    cash_paid_to_financing_activities = "cash_paid_to_financing_activities"
    fcff = "fcff"
    cash_paid_for_employee = "cash_paid_for_employee"
    cash_paid_for_goods_and_services = "cash_paid_for_goods_and_services"
    cash_from_other_operating_activities = "cash_from_other_operating_activities"
    cash_received_from_disposal_of_asset = "cash_received_from_disposal_of_asset"
    cash_received_from_sales_of_goods = "cash_received_from_sales_of_goods"
    deferred_expense_amortization = "deferred_expense_amortization"
    cash_paid_for_other_investment_activities = "cash_paid_for_other_investment_activities"
    fcfe = "fcfe"
    cash_flow_from_financing_activities = "cash_flow_from_financing_activities"
    cash_received_from_other_financing_activities = "cash_received_from_other_financing_activities"
    cash_paid_for_dividend_and_interest = "cash_paid_for_dividend_and_interest"
    cash_paid_for_debt = "cash_paid_for_debt"
    fixed_asset_depreciation = "fixed_asset_depreciation"
    cash_paid_for_investment_activities = "cash_paid_for_investment_activities"
    cash_paid_for_asset = "cash_paid_for_asset"
    cash_received_from_financing_activities = "cash_received_from_financing_activities"
    cash_flow_from_investing_activities = "cash_flow_from_investing_activities"
    cash_received_from_investment_activities = "cash_received_from_investment_activities"
    net_inc_cash_and_equivalents = "net_inc_cash_and_equivalents"
    intangible_asset_amortization = "intangible_asset_amortization"
    depreciation_and_amortization = "depreciation_and_amortization"
    cash_paid_for_other_operation_activities = "cash_paid_for_other_operation_activities"
    cash_received_from_financial_institution_borrows = "cash_received_from_financial_institution_borrows"
    cash_received_from_other_investment_activities = "cash_received_from_other_investment_activities"
    cash_paid_to_acquire_investment = "cash_paid_to_acquire_investment"
    cash_flow_from_operating_activities = "cash_flow_from_operating_activities"
    cash_equivalent_increase = "cash_equivalent_increase"
    cash_paid_for_operation_activities = "cash_paid_for_operation_activities"
    cash_paid_for_taxes = "cash_paid_for_taxes"
    refunds_of_taxes = "refunds_of_taxes"
    assets_depreciation_reserves = "assets_depreciation_reserves"
    net_cash_deal_from_sub = "net_cash_deal_from_sub"
    net_cash_payment_from_sub = "net_cash_payment_from_sub"


class IncomeStatement(PitBase):
    total_income_minority = "total_income_minority"
    ga_expense = "ga_expense"
    invest_income_associates = "invest_income_associates"
    profit_from_operation = "profit_from_operation"
    r_n_d = "r_n_d"
    total_income = "total_income"
    other_operating_revenue = "other_operating_revenue"
    total_income_parent_company = "total_income_parent_company"
    other_operating_cost = "other_operating_cost"
    revenue = "revenue"
    net_profit_parent_company = "net_profit_parent_company"
    non_operating_revenue = "non_operating_revenue"
    selling_expense = "selling_expense"
    ebitda = "ebitda"
    minority_profit = "minority_profit"
    disposal_loss_on_asset = "disposal_loss_on_asset"
    non_recurring_pnl = "non_recurring_pnl"
    exchange_gains_or_losses = "exchange_gains_or_losses"
    financing_expense = "financing_expense"
    interest_expense = "interest_expense"
    operating_revenue = "operating_revenue"
    net_profit = "net_profit"
    total_expense = "total_expense"
    fair_value_change_income = "fair_value_change_income"
    other_income = "other_income"
    asset_impairment = "asset_impairment"
    non_operating_expense = "non_operating_expense"
    net_profit_deduct_non_recurring_pnl = "net_profit_deduct_non_recurring_pnl"
    profit_before_tax = "profit_before_tax"
    sales_tax = "sales_tax"
    basic_earnings_per_share = "basic_earnings_per_share"
    ebit = "ebit"
    investment_income = "investment_income"
    income_tax = "income_tax"
    gross_profit = "gross_profit"
    cost_of_goods_sold = "cost_of_goods_sold"
    interest_income = "interest_income"
    unrealised_investment_loss = "unrealised_investment_loss"
    fully_diluted_earnings_per_share = "fully_diluted_earnings_per_share"
    operating_expense = "operating_expense"
    o_n_a_expense = "o_n_a_expense"
    gredit_asset_impairment = "gredit_asset_impairment"
    disposal_income_on_asset = "disposal_income_on_asset"
    financing_interest_income = "financing_interest_income"
    financing_interest_expense = "financing_interest_expense"


class TTM(PitBase):
    operating_expenseTTM = "operating_expenseTTM"
    operating_costTTM = "operating_costTTM"
    net_invest_cashflowTTM = "net_invest_cashflowTTM"
    sale_service_render_cashTTM = "sale_service_render_cashTTM"
    ebitTTM = "ebitTTM"
    operating_revenueTTM = "operating_revenueTTM"
    operating_profitTTM = "operating_profitTTM"
    ni_from_operatingTTM = "ni_from_operatingTTM"
    net_cashflowTTM = "net_cashflowTTM"
    net_finance_cashflowTTM = "net_finance_cashflowTTM"
    net_operate_cashflowTTM = "net_operate_cashflowTTM"
    operating_payoutTTM = "operating_payoutTTM"
    asset_impairment_lossTTM = "asset_impairment_lossTTM"
    non_operating_net_incomeTTM = "non_operating_net_incomeTTM"
    administration_expenseTTM = "administration_expenseTTM"
    gross_profitTTM = "gross_profitTTM"
    total_operating_costTTM = "total_operating_costTTM"
    financial_expenseTTM = "financial_expenseTTM"
    ni_from_value_changeTTM = "ni_from_value_changeTTM"
    np_parent_company_ownersTTM = "np_parent_company_ownersTTM"
    net_profitTTM = "net_profitTTM"
    total_profitTTM = "total_profitTTM"
    total_operating_revenueTTM = "total_operating_revenueTTM"


class MainData(PitBase):
    adjusted_net_profit = "adjusted_net_profit"


class RndExpenditureSheet(PitBase):
    expensed_rnd = "expensed_rnd"
    capitalized_rnd = "capitalized_rnd"
    total_rnd = "total_rnd"
    rnd_to_revenue = "rnd_to_revenue"
    capitalized_rnd_ratio = "capitalized_rnd_ratio"
    rnd_employees = "rnd_employees"
    rnd_employees_ratio = "rnd_employees_ratio"


PIT_TABLES = ('BalanceSheet', 'CashFlowsStatement', 'IncomeStatement', 'TTM', 'MainData')
PIT_FIELDS = (
    {v for k, v in BalanceSheet.__dict__.items() if not k.startswith('_')},
    {v for k, v in CashFlowsStatement.__dict__.items() if not k.startswith('_')},
    {v for k, v in IncomeStatement.__dict__.items() if not k.startswith('_')},
    {v for k, v in TTM.__dict__.items() if not k.startswith('_')},
    {v for k, v in MainData.__dict__.items() if not k.startswith('_')},
    {v for k, v in RndExpenditureSheet.__dict__.items() if not k.startswith('_')},
)
PIT_BASIC_FIELDS = ("order_book_id", "end_date", "if_adjusted", "info_date", "accounting_standards",
                    "info_type", "if_complete", "enterprise_type")
PIT_BASIC_FIELDS_SET = set(PIT_BASIC_FIELDS)



# -*- coding: utf-8 -*-
import logging
import time
import types
import warnings
from functools import wraps, partial, update_wrapper

import rqdatac
from rqdatac.share.errors import OverwriteWarning

log = logging.getLogger("rqdata")


def deprecated(func=None, msg=None):
    if func is None:
        return partial(deprecated, msg=msg)

    if msg is None:
        msg = func.__name__ + " is deprecated, and will be removed in future."

    @wraps(func)
    def wrap(*args, **kwargs):
        warnings.warn(msg, category=FutureWarning, stacklevel=0)
        return func(*args, **kwargs)

    return wrap


def export_as_api(f=None, name=None, namespace=None, priority=0):
    if f is None:
        return partial(export_as_api, name=name, namespace=namespace)
    name = name if name else f.__name__
    if namespace:
        if hasattr(rqdatac, namespace):
            namespace = getattr(rqdatac, namespace)
        else:
            namespace_name = namespace
            namespace = types.ModuleType(namespace_name)
            namespace.__file__ = 'rqdatac plugin'
            namespace.__module__ = "rqdatac"
            setattr(rqdatac, namespace_name, namespace)
            rqdatac.__all__.append(namespace_name)
    else:
        namespace = rqdatac
        rqdatac.__all__.append(name)

    old_f = getattr(namespace, name, None)
    if old_f is not None:
        if old_f.__priority > priority:
            warnings.warn("!!!! CAN'T OVERWRITE API {} WITH {} BECAUSE OLD PRIPORITY {} > {} !!!!".format(name, f, old_f.__priority, priority), category=OverwriteWarning)
            return f
        warnings.warn("!!!! OVERWRITE API {} WITH {} !!!!".format(name, f), category=OverwriteWarning)

    f.__priority = priority
    setattr(namespace, name, f)

    return f


def retry(count, suppress_exceptions, delay=1.0):
    def decorate(func):
        @wraps(func)
        def wrap(*args, **kwargs):
            c = count
            while c > 0:
                try:
                    return func(*args, **kwargs)
                except suppress_exceptions as e:
                    c -= 1
                    if c == 0:
                        raise e
                    log.error("Retry because %r", e)
                    if delay:
                        time.sleep(delay)

        return wrap

    return decorate


def coroutine(func):
    @wraps(func)
    def primer(*args, **kwargs):
        gen = func(*args, **kwargs)
        next(gen)
        return gen

    return primer


_ttl_cached_functions = set()


def ttl_cache(ttl):
    if not isinstance(ttl, int) or not ttl > 0:
        raise TypeError("Expected ttl to be a positive integer")

    def decorating_function(user_function):
        wrapper = _ttl_cache_wrapper(user_function, ttl)
        _ttl_cached_functions.add(wrapper)
        return update_wrapper(wrapper, user_function)

    return decorating_function


def ttl_cache_clear():
    for f in _ttl_cached_functions:
        f.clear()


def _ttl_cache_wrapper(user_function, ttl):
    sentinel = object()
    cache = {}
    cache_get = cache.get  # bound method to lookup a key or return None
    cache_clear = cache.clear

    def wrapper(*args, **kwargs):
        if kwargs:
            key = args + (repr(sorted(kwargs.items())),)
        else:
            key = args

        # in cpython, dict.get is thread-safe
        result = cache_get(key, sentinel)
        if result is not sentinel:
            expire_at, value = result
            if expire_at > time.time():
                return value
        value = user_function(*args, **kwargs)
        cache[key] = (time.time() + ttl, value)
        return value

    setattr(wrapper, 'clear', cache_clear)
    return wrapper


def compatible_with_parm(func=None, name=None, value=None, replace=None):
    if func is None:
        return partial(compatible_with_parm, name=name, value=value, replace=replace)

    @wraps(func)
    def wrap(*args, **kwargs):
        if name:
            if name in kwargs:
                msg = "'{}' is no longer used, please use '{}' instead ".format(name, replace)
                warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
                item = kwargs.pop(name)
                if item != value:
                    raise ValueError("'{}': except '{}', got '{}'".format(name, value, item))
        return func(*args, **kwargs)

    return wrap



# -*- coding: utf-8 -*-
from sqlalchemy import Numeric, Column
from .fundamental_base_sql import FundamentalBase


class CashFlowStatementTTM(FundamentalBase):
    net_finance_cashflowTTM = Column(Numeric(18, 4))
    net_operate_cashflowTTM = Column(Numeric(18, 4))
    net_invest_cashflowTTM = Column(Numeric(18, 4))
    net_cashflowTTM = Column(Numeric(18, 4))
    sale_service_render_cashTTM = Column(Numeric(18, 4))


class IncomeStatementTTM(FundamentalBase):
    np_parent_company_ownersTTM = Column(Numeric(18, 4))
    financial_expenseTTM = Column(Numeric(18, 4))
    administration_expenseTTM = Column(Numeric(18, 4))
    ni_from_value_changeTTM = Column(Numeric(18, 4))
    ni_from_operatingTTM = Column(Numeric(18, 4))
    net_profitTTM = Column(Numeric(18, 4))
    total_profitTTM = Column(Numeric(18, 4))
    gross_profitTTM = Column(Numeric(18, 4))
    ebitTTM = Column(Numeric(18, 4))
    operating_expenseTTM = Column(Numeric(18, 4))
    operating_costTTM = Column(Numeric(18, 4))
    operating_profitTTM = Column(Numeric(18, 4))
    operating_revenueTTM = Column(Numeric(18, 4))
    non_operating_net_incomeTTM = Column(Numeric(18, 4))
    operating_payoutTTM = Column(Numeric(18, 4))
    total_operating_costTTM = Column(Numeric(18, 4))
    total_operating_revenueTTM = Column(Numeric(18, 4))
    asset_impairment_lossTTM = Column(Numeric(18, 4))


class FinancialIndicatorTTM(FundamentalBase):
    eps_belongs_to_parentTTM = Column(Numeric(18, 4))
    epsTTM = Column(Numeric(18, 4))
    financial_expense_rateTTM = Column(Numeric(18, 4))
    invest_associates_inc_to_tpTTM = Column(Numeric(18, 4))
    admin_expense_rateTTM = Column(Numeric(18, 4))
    value_change_ni_to_tpTTM = Column(Numeric(18, 4))
    no_cf_to_operating_niTTM = Column(Numeric(18, 4))
    cash_rate_of_salesTTM = Column(Numeric(18, 4))
    operating_mi_to_tpTTM = Column(Numeric(18, 4))
    net_profit_to_total_operating_revenueTTM = Column(Numeric(18, 4))
    return_on_equityTTM = Column(Numeric(18, 4))
    operating_cashflow_per_shareTTM = Column(Numeric(18, 4))
    cashflow_per_shareTTM = Column(Numeric(18, 4))
    operating_revenue_per_shareTTM = Column(Numeric(18, 4))
    ebit_to_total_revenueTTM = Column(Numeric(18, 4))
    operating_expense_rateTTM = Column(Numeric(18, 4))
    net_profit_ratioTTM = Column(Numeric(18, 4))
    gross_income_ratioTTM = Column(Numeric(18, 4))
    period_costs_rateTTM = Column(Numeric(18, 4))
    sale_service_cash_to_operating_revenueTTM = Column(Numeric(18, 4))
    operating_profit_to_total_revenueTTM = Column(Numeric(18, 4))
    net_non_operating_income_to_revenueTTM = Column(Numeric(18, 4))
    total_operating_cost_to_total_operating_revenueTTM = Column(Numeric(18, 4))
    asset_impair_loss_to_total_operating_revenueTTM = Column(Numeric(18, 4))
    return_on_asset_ebitTTM = Column(Numeric(18, 4))
    return_on_assetTTM = Column(Numeric(18, 4))



# -*- coding: utf-8 -*-
import datetime
import pandas as pd
from rqdatac.client import get_client
from rqdatac.validators import ensure_string, check_items_in_container, ensure_order_book_id
from rqdatac.utils import to_datetime
from rqdatac.decorators import export_as_api


class Tmall:
    @staticmethod
    def stocks():
        """获取电商天猫数据股票列表

        :return:
            股票列表
        """

        return get_client().execute("tmall.stocks")

    @staticmethod
    def data(order_book_id, start_date=None, end_date=None, frequency="1d", fields=None):
        """获取天猫电商销售额数据

        :param order_book_id: 股票名
        :param start_date: 开始日期，默认为结束日期前一个月,  必须在2016年6月30日之后 (Default value = None)
        :param end_date: 结束日期 (Default value = None)
        :param frequency: 如'1d', '1M' (Default value = "1d")
        :param fields: 如'sales' (Default value = None)
        :returns: 返回DataFrame

        """
        order_book_id = ensure_order_book_id(order_book_id)

        if not end_date:
            end_date = datetime.date.today()

        if not start_date:
            start_date = end_date - datetime.timedelta(days=30)

        end_date = to_datetime(end_date)
        start_date = to_datetime(start_date)

        if start_date < datetime.datetime(2016, 6, 30):
            raise ValueError("start_date cannot be earlier than 2016-06-30")

        if start_date > end_date:
            raise ValueError()

        ensure_string(frequency, "frequency")
        check_items_in_container([frequency], {"1d", "1M"}, "frequency")

        if fields is None:
            fields = "sales"
        else:
            fields = ensure_string(fields, "fields")
            check_items_in_container(fields, ["sales"], "fields")
        data = get_client().execute("tmall.data", order_book_id, start_date, end_date, frequency)
        if not data:
            return
        df = pd.DataFrame(data)
        df = df.set_index("date")
        df.sort_index(inplace=True)
        df.columns = ["sales"]
        return df


@export_as_api
class ecommerce:
    tmall = Tmall



# -*- coding: utf-8 -*-

"""
    2 bytes    |1 byte |1 byte |            4 bytes            |
----------------------------------------------------------------
|   Msg Type   |   ST  |   CM  |           Body Length         |
----------------------------------------------------------------

Msg Type: 消息类型
ST: Serialization Type, 序列化方式
CM: Compression Method, 压缩方式
Body Length: 包体长度，不包含头

little-endian
"""


def with_metaclass(mcls):
    def decorator(cls):
        body = vars(cls).copy()
        # clean out class body
        body.pop("__dict__", None)
        body.pop("__weakref__", None)
        return mcls(cls.__name__, cls.__bases__, body)

    return decorator


class SimpleEnum(type):
    def __new__(mcs, name, bases, attrs):
        mapping = {v: k for k, v in attrs.items() if not k.startswith("_")}
        attrs["__map"] = mapping
        attrs["__key"] = frozenset(mapping.keys())
        return type.__new__(mcs, name, bases, attrs)

    def __getitem__(self, item):
        _map = self.__dict__["__map"]
        if item not in self.__dict__["__key"]:
            return "unknown %s(%s)" % (self.__name__, item)
        return _map[item]

    def __contains__(self, item):
        return item in self.__dict__["__key"]


@with_metaclass(SimpleEnum)
class MSG_TYPE:
    HANDSHAKE = 0
    REQUEST = 1
    RESPONSE = 2
    ERROR = 3
    STREAM_START = 4
    STREAM_END = 5
    FEED = 6
    TABLE = 7
    TABLE2 = 8


@with_metaclass(SimpleEnum)
class SERIALIZATION_TYPE:
    RAW = 0
    JSON = 1
    MSGPACK = 2


@with_metaclass(SimpleEnum)
class COMPRESSION_METHOD:
    NONE = 0
    SNAPPY = 1
    ZLIB = 2
    BROTLI = 3


HEADER_FORMAT = "<HBBI"
HEADER_LENGTH = 8



# -*- coding: utf-8 -*-
import datetime

import pandas as pd
import numpy as np

from rqdatac.services.calendar import get_previous_trading_date
from rqdatac.validators import (
    ensure_string_in,
    ensure_order_book_id,
    ensure_order_book_ids,
    ensure_date_range,
)
from rqdatac.utils import (
    int8_to_datetime_v,
    int14_to_datetime_v,
    int17_to_datetime_v,
    int17_to_datetime,
    today_int,
    date_to_int8,
)
from rqdatac.client import get_client
from rqdatac.decorators import export_as_api
from rqdatac.share.errors import MarketNotSupportError, PermissionDenied

DAYBAR_FIELDS = [
    "close", "volume", "total_turnover"
]
TICKBAR_FIELDS = [
    "datetime", "close", "volume", "total_turnover", "bid_vol", "ask_vol"
]


def convert_bar_to_multi_df(data, dt_name, fields, convert_dt):
    line_no = 0
    dt_set = set()
    obid_level = []
    obid_slice_map = {}
    for obid, d in data:
        dts = d[dt_name]
        dts_len = len(dts)
        if dts_len == 0:
            continue
        obid_slice_map[obid] = slice(line_no, line_no + dts_len, None)
        dt_set.update(dts)
        line_no += dts_len

        obid_level.append(obid)

    if line_no == 0:
        return

    obid_idx_map = {o: i for i, o in enumerate(obid_level)}
    obid_label = np.empty(line_no, dtype=object)
    dt_label = np.empty(line_no, dtype=object)
    arr = np.full((line_no, len(fields)), np.nan)
    r_map_fields = {f: i for i, f in enumerate(fields)}

    dt_arr_sorted = np.array(sorted(dt_set))
    dt_level = convert_dt(dt_arr_sorted)

    for obid, d in data:
        dts = d[dt_name]
        if len(dts) == 0:
            continue
        slice_ = obid_slice_map[obid]
        for f, value in d.items():
            if f == dt_name:
                dt_label[slice_] = dt_arr_sorted.searchsorted(dts, side='left')
            else:
                arr[slice_, r_map_fields[f]] = value
        obid_label[slice_] = [obid_idx_map[obid]] * len(dts)
    try:
        # func 'is_datetime_with_singletz_array'  is the most time consuming part in multi_index constructing
        # it is useless for our situation. skip it.
        func_is_singletz = getattr(pd._libs.lib, 'is_datetime_with_singletz_array')
        setattr(pd._libs.lib, 'is_datetime_with_singletz_array', lambda *args: True)
    except AttributeError:
        func_is_singletz = None

    multi_idx = pd.MultiIndex([obid_level, dt_level], [obid_label, dt_label],
                              names=('order_book_id', dt_name))

    if func_is_singletz is not None:
        # recovery
        setattr(pd._libs.lib, 'is_datetime_with_singletz_array', func_is_singletz)

    df = pd.DataFrame(data=arr, index=multi_idx, columns=fields)
    return df


def get_auction_info_daybar(order_book_ids, start_date, end_date, fields, duration=1, market="cn"):
    data = get_client().execute(
        "get_auction_info_daybar", order_book_ids, start_date, end_date, fields, duration, market
    )
    data = [(obid, {k: np.frombuffer(*v) for k, v in d.items()}) for obid, d in data]
    res = convert_bar_to_multi_df(data, 'date', fields, int8_to_datetime_v)
    return res


def get_today_auction_info_minbar(order_book_ids, date, fields, duration, market="cn"):
    data = get_client().execute("get_today_auction_info_minbar", order_book_ids, date, fields, duration, market)
    return convert_bar_to_multi_df(data, "datetime", fields, int14_to_datetime_v)


def get_auction_info_minbar(order_book_ids, start_date, end_date, fields, duration, market):
    data = get_client().execute(
        "get_auction_info_minbar", order_book_ids, start_date, end_date, fields, duration, market
    )
    if data:
        data = [(obid, {k: np.frombuffer(*v) for k, v in d.items()}) for obid, d in data]
        df = convert_bar_to_multi_df(data, 'datetime', fields, int14_to_datetime_v)
    else:
        df = None

    today = today_int()
    if df is None:
        history_latest_date = date_to_int8(get_previous_trading_date(today, market=market))
    else:
        history_latest_date = date_to_int8(df.index.get_level_values(1).max())

    if history_latest_date >= end_date or start_date > today or history_latest_date >= today:
        return df
    try:
        live_df = get_today_auction_info_minbar(order_book_ids, today, fields, duration, market)
    except (MarketNotSupportError, PermissionDenied):
        live_df = None
    if live_df is None:
        return df
    if df is None:
        return live_df
    df = pd.concat([df, live_df])
    df.sort_index(inplace=True)
    return df


def get_today_auction_info_tick(order_book_id, date, fields, market="cn"):
    data = get_client().execute("get_today_auction_info_tick", order_book_id, date, market)
    df = pd.DataFrame(data[0])
    if df.empty:
        return None
    df = df[fields]
    df.datetime = df.datetime.apply(int17_to_datetime)
    df.set_index("datetime", inplace=True)
    return df


def get_auction_info_tickbar(order_book_id, start_date, end_date, fields, market):
    order_book_id = ensure_order_book_id(order_book_id)
    start_date, end_date = ensure_date_range(start_date, end_date, datetime.timedelta(days=3))
    data = get_client().execute(
        "get_auction_info_tickbar", order_book_id, start_date, end_date, fields, market
    )
    today = today_int()
    if data:
        data = [(obid, {k: np.frombuffer(*v) for k, v in d.items()}) for obid, d in data]
        df_list = []
        for obid, d in data:
            df = pd.DataFrame(d)
            df_list.append(df)

        df = pd.concat(df_list)  # type: pd.DataFrame
        df["datetime"] = int17_to_datetime_v(df["datetime"].values)
        history_latest_date = date_to_int8(df.iloc[-1]["datetime"])
        df.set_index("datetime", inplace=True)
    else:
        df = None
        history_latest_date = date_to_int8(get_previous_trading_date(today, market=market))

    if history_latest_date >= end_date or start_date > today or history_latest_date >= today:
        return df
    try:
        live_df = get_today_auction_info_tick(order_book_id, today, fields, market=market)
    except (MarketNotSupportError, PermissionDenied):
        live_df = None
    if live_df is None:
        return df
    if df is None:
        return live_df
    return pd.concat([df, live_df])


@export_as_api
def get_ksh_auction_info(order_book_ids, start_date=None, end_date=None, frequency="1d", market="cn"):
    import warnings

    msg = "'get_ksh_auction_info' is deprecated, and will be removed on 2021-01-29, " \
          "use 'get_auction_info' instead."
    warnings.warn(msg, stacklevel=2)
    return get_auction_info(order_book_ids, start_date, end_date, frequency, market)


@export_as_api
def get_auction_info(order_book_ids, start_date=None, end_date=None, frequency="1d", market="cn"):
    """获取股票盘后数据
    :param order_book_ids: 股票代码or股票代码列表, 如'000001.XSHE'
    :param start_date: 开始日期
    :param end_date: 结束日期
    :param frequency: 默认为日线。日线使用 '1d', 分钟线 '1m'  快照 'tick' (Default value = "1d"),
    :param market:  (Default value = "cn")
    :returns: pandas.DataFrame or None
    """
    ensure_string_in(frequency, ("1d", "1m", "tick"), "frequency")
    if frequency == "tick":
        return get_auction_info_tickbar(order_book_ids, start_date, end_date, TICKBAR_FIELDS, market)

    order_book_ids = ensure_order_book_ids(order_book_ids)
    start_date, end_date = ensure_date_range(start_date, end_date)
    if frequency == "1d":
        return get_auction_info_daybar(order_book_ids, start_date, end_date, DAYBAR_FIELDS, 1, market)

    return get_auction_info_minbar(order_book_ids, start_date, end_date, DAYBAR_FIELDS, 1, market)